from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from webviz_config import WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from webviz_subsurface._models import (
    EnsembleSetModel,
    caching_ensemble_set_model_factory,
)
from webviz_subsurface._models.parameter_model import ParametersModel
from webviz_subsurface._utils.unique_theming import unique_colors


class RftPlotterDataModel:
    """Class keeping the data needed in the RFT vizualisations and various
    data providing methods.
    """

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: Optional[List[str]],
        formations: Path = None,
        faultlines: Path = None,
        obsdata: Path = None,
        csvfile_rft: Path = None,
        csvfile_rft_ert: Path = None,
    ):
        self.formations = formations
        self.faultlines = faultlines
        self.obsdata = obsdata
        self.csvfile_rft = csvfile_rft
        self.csvfile_rft_ert = csvfile_rft_ert

        if csvfile_rft_ert and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile_rft_ert" or "ensembles"'
            )

        self.simdf = read_csv(self.csvfile_rft) if csvfile_rft is not None else None
        self.formationdf = read_csv(self.formations) if self.formations else None
        self.faultlinesdf = read_csv(self.faultlines) if self.faultlines else None
        self.obsdatadf = read_csv(self.obsdata) if self.obsdata else None
        self.ertdatadf = pd.DataFrame()

        if csvfile_rft_ert is not None:
            self.ertdatadf = read_csv(self.csvfile_rft_ert)

            # Must send a dummy dataframe to ParametersModel
            # The ensembles will be identified as sensitivity runs
            self.param_model = ParametersModel(
                pd.DataFrame(
                    columns=["REAL", "ENSEMBLE", "SENSNAME", "SENSCASE"],
                    data=[
                        [0, "ensemble", "sensname", "low"],
                    ],
                )
            )

        if ensembles is not None:
            ens_paths = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.emodel: EnsembleSetModel = (
                caching_ensemble_set_model_factory.get_or_create_model(
                    ensemble_paths=ens_paths,
                )
            )
            try:
                self.simdf = self.emodel.load_csv(Path("share/results/tables/rft.csv"))
            except (KeyError, OSError):
                self.simdf = None

            self.param_model = ParametersModel(
                dataframe=self.emodel.load_parameters(),
                drop_constants=True,
                keep_numeric_only=True,
            )

            try:
                self.ertdatadf = self.emodel.load_csv(
                    Path("share/results/tables/rft_ert.csv")
                )
            except KeyError as exc:
                raise KeyError(
                    "CSV file for ERT RFT observations/simulations "
                    "(share/results/tables/rft_ert.csv) not found!"
                ) from exc

        self.ertdatadf = self.ertdatadf.rename(
            columns={
                "time": "DATE",
                "is_active": "ACTIVE",
                "isactive": "ACTIVE",
                "well": "WELL",
                "zone": "ZONE",
                "pressure": "SIMULATED",
                "true_vertical_depth": "TVD",
                "measured_depth": "MD",
                "observed": "OBSERVED",
                "obs": "OBSERVED",
                "error": "OBSERVED_ERR",
                "utm_x": "EAST",
                "utm_y": "NORTH",
            }
        )
        self.ertdatadf["DIFF"] = (
            self.ertdatadf["SIMULATED"] - self.ertdatadf["OBSERVED"]
        )
        self.ertdatadf["ABSDIFF"] = abs(
            self.ertdatadf["SIMULATED"] - self.ertdatadf["OBSERVED"]
        )
        self.ertdatadf["YEAR"] = pd.to_datetime(self.ertdatadf["DATE"]).dt.year
        self.ertdatadf = self.ertdatadf.sort_values(by="DATE")
        self.ertdatadf["DATE_IDX"] = self.ertdatadf["DATE"].apply(
            lambda x: list(self.ertdatadf["DATE"].unique()).index(x)
        )
        self.date_marks = self.set_date_marks()
        self.ertdatadf = filter_frame(
            self.ertdatadf,
            {
                "ACTIVE": 1,
            },
        )
        self.ertdatadf["STDDEV"] = self.ertdatadf.groupby(
            ["WELL", "DATE", "ZONE", "ENSEMBLE", "TVD"]
        )["SIMULATED"].transform("std")

    @property
    def well_names(self) -> List[str]:
        return sorted(list(self.ertdatadf["WELL"].unique()))

    @property
    def zone_names(self) -> List[str]:
        return sorted(list(self.ertdatadf["ZONE"].unique()))

    @property
    def dates(self) -> List[str]:
        return sorted(list(self.ertdatadf["DATE"].unique()))

    def date_in_well(self, well: str) -> List[str]:
        df = self.ertdatadf.loc[self.ertdatadf["WELL"] == well]
        return [str(d) for d in list(df["DATE"].unique())]

    def well_dates_and_zones(self, well: str) -> Tuple[List[str], List[str]]:
        df = self.ertdatadf.loc[self.ertdatadf["WELL"] == well]
        return [str(d) for d in list(df["DATE"].unique())], list(df["ZONE"].unique())

    @property
    def ensembles(self) -> List[str]:
        return list(self.ertdatadf["ENSEMBLE"].unique())

    @property
    def enscolors(self) -> dict:
        return unique_colors(self.ensembles)

    @property
    def parameters(self) -> List[str]:
        return self.param_model.parameters

    def set_date_marks(self) -> Dict[str, Dict[str, Any]]:
        marks = {}
        idx_steps = np.linspace(
            start=0,
            stop=self.ertdatadf["DATE_IDX"].max(),
            num=min(4, len(self.ertdatadf["DATE_IDX"].unique())),
            dtype=int,
        )
        date_steps = self.ertdatadf.loc[self.ertdatadf["DATE_IDX"].isin(idx_steps)][
            "DATE"
        ].unique()

        for i, date_index in enumerate(idx_steps):
            marks[str(date_index)] = {
                "label": f"{date_steps[i]}",
                "style": {
                    "white-space": "nowrap",
                    "font-weight": "bold",
                },
            }
        return marks

    def get_param_real_and_value_df(
        self, ensemble: str, parameter: str, normalize: bool = False
    ) -> pd.DataFrame:
        """
        Return dataframe with ralization and values for selected parameter for an ensemble.
        A column with normalized parameter values can be added.
        """
        df = self.param_model.dataframe.melt(
            id_vars=["ENSEMBLE", "REAL"], var_name="PARAMETER", value_name="VALUE"
        )
        df = df[["VALUE", "REAL"]].loc[
            (df["ENSEMBLE"] == ensemble) & (df["PARAMETER"] == parameter)
        ]
        if normalize:
            df["VALUE_NORM"] = (df["VALUE"] - df["VALUE"].min()) / (
                df["VALUE"].max() - df["VALUE"].min()
            )
        return df.reset_index(drop=True)

    def create_rft_and_param_pivot_table(
        self,
        ensemble: str,
        well: str,
        date: str,
        zone: str,
        reals: List[int],
        keep_all_rfts: bool,
    ) -> Tuple[Optional[pd.DataFrame], float, float, List[str], List[str],]:
        """This method merges rft observations and parameters.

        The RFT observations are converted from long form (all simulated values
        in one column: SIMULATED) to wide form (on column for each well/date/zone).
        This is needed for calculating correlations between RFTs. It is an option
        whether to keep one or all RFT columns. If there are multiple observations
        with the same well/date/zone, they are averaged (depth could be added as a
        fourth parameter here, f.ex optional).

        This RFT wide form table is then merged on REAL with the ensemble parameters.

        Returns:
        * merged dataframe with RFTs and parameters
        * observation for well/date/zone
        * observation error
        * list with ensemble parameters
        * list with rft names
        """
        # pylint: disable = too-many-locals
        rft_filter = {"ENSEMBLE": ensemble, "REAL": reals}
        if not keep_all_rfts:
            rft_filter.update({"WELL": well, "DATE": date, "ZONE": zone})

        rft_df = filter_frame(self.ertdatadf, rft_filter)[
            ["REAL", "DATE", "WELL", "ZONE", "SIMULATED", "OBSERVED", "OBSERVED_ERR"]
        ]

        if rft_df.empty:
            return None, 0, 0, [], []

        rft_df["RFT_KEY"] = rft_df["WELL"] + " " + rft_df["DATE"] + " " + rft_df["ZONE"]

        # Get observation and error
        # In case there are multiple observations in the same well/date/zone
        # they are averaged (as in the pivot table)
        df_current_obs = rft_df[rft_df["RFT_KEY"] == f"{well} {date} {zone}"]
        obs = df_current_obs["OBSERVED"].mean()
        obs_err = df_current_obs["OBSERVED_ERR"].mean()

        pivot_df = rft_df.pivot_table(
            index="REAL", columns="RFT_KEY", values="SIMULATED", aggfunc="mean"
        ).reset_index()

        param_df = (
            filter_frame(
                self.param_model.dataframe, {"ENSEMBLE": ensemble, "REAL": reals}
            )
            .drop("ENSEMBLE", axis=1)
            .dropna(axis=1)  # Removes parameters not used in this ensemble
        )
        ens_params = [
            param
            for param in param_df.columns
            if param not in self.param_model.POSSIBLE_SELECTORS
        ]
        ens_rfts = [rft for rft in pivot_df.columns if rft != "REAL"]

        return pivot_df.merge(param_df, on="REAL"), obs, obs_err, ens_params, ens_rfts

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[Dict[str, Any]]]]:
        functions: List[Tuple[Callable, List[Dict[str, Any]]]] = [
            (
                read_csv,
                [
                    {"csv_file": path}
                    for path in [
                        self.faultlines,
                        self.formations,
                        self.obsdata,
                        self.csvfile_rft,
                        self.csvfile_rft_ert,
                    ]
                    if path is not None
                ],
            )
        ]
        if self.csvfile_rft_ert is None:
            functions.extend(self.emodel.webvizstore)
        return functions


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: str) -> pd.DataFrame:
    return pd.read_csv(csv_file)


def interpolate_depth(df: pd.DataFrame) -> pd.DataFrame:
    df = (
        df.pivot_table(index=["DEPTH"], columns=["REAL"], values="PRESSURE")
        .interpolate(limit_direction="both")
        .stack("REAL")
    )
    return df.to_frame().rename(columns={0: "PRESSURE"}).reset_index()


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def filter_frame(
    dframe: pd.DataFrame, column_values: Dict[str, Union[List[str], str]]
) -> pd.DataFrame:
    """General function for filtering dataframes"""
    df = dframe.copy()
    for column, value in column_values.items():
        if isinstance(value, list):
            df = df.loc[df[column].isin(value)]
        else:
            df = df.loc[df[column] == value]
    return df


def correlate(df: pd.DataFrame, response: str) -> pd.Series:
    """Returns the correlation matrix for a dataframe

    This function is the same as in ParameterAnalysis and could be generalized
    """
    df = df[df.columns[df.nunique() > 1]].copy()
    if response not in df.columns:
        df[response] = np.nan
    series = df[response]
    df = df.drop(columns=[response])
    corrdf = df.corrwith(series)
    corrdf.fillna(0, inplace=True)
    return corrdf.reindex(corrdf.abs().sort_values().index)



from enum import Enum


class MapAttribute(Enum):
    MIGRATION_TIME = "Migration Time"
    MAX_SGAS = "Maximum SGAS"
    MAX_AMFG = "Maximum AMFG"
    SGAS_PLUME = "Plume (SGAS)"
    AMFG_PLUME = "Plume (AMFG)"



import glob
import io
import json
import logging
import re
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import pyarrow as pa
from ecl2df.vfp import pyarrow2basic_data
from ecl2df.vfp._vfpdefs import (
    ALQ,
    GFR,
    THPTYPE,
    UNITTYPE,
    VFPPROD_FLO,
    VFPPROD_TABTYPE,
    VFPPROD_UNITS,
    VFPTYPE,
    WFR,
)
from webviz_config import WebvizSettings
from webviz_config.webviz_store import webvizstore

from .._types import PressureType, VfpParam


class VfpTable:
    """Class that contains data and metadata for one VFP table"""

    def __init__(self, filename: str):
        self._filename = filename
        self._data = json.load(_read_vfp_arrow(self._filename))
        self._vfp_type = VFPTYPE(self._data["VFP_TYPE"])
        if self._vfp_type == VFPTYPE.VFPINJ:
            raise NotImplementedError(
                f"""
Could not load {self._filename}. VFPINJ tables not implemented.
            """
            )

        self._table_number = self._data["TABLE_NUMBER"]
        self._tab_type = VFPPROD_TABTYPE(self._data["TAB_TYPE"])
        self._unit_type = UNITTYPE(self._data["UNIT_TYPE"])
        self._datum = self._data["DATUM"]

        self.params = {
            VfpParam.THP: dict(enumerate(self._data["THP_VALUES"])),
            VfpParam.WFR: dict(enumerate(self._data["WFR_VALUES"])),
            VfpParam.GFR: dict(enumerate(self._data["GFR_VALUES"])),
            VfpParam.ALQ: dict(enumerate(self._data["ALQ_VALUES"])),
            VfpParam.RATE: dict(enumerate(self._data["FLOW_VALUES"])),
        }
        self.param_types = {
            VfpParam.THP: THPTYPE(self._data["THP_TYPE"]),
            VfpParam.WFR: WFR(self._data["WFR_TYPE"]),
            VfpParam.GFR: GFR(self._data["GFR_TYPE"]),
            VfpParam.ALQ: ALQ(self._data["ALQ_TYPE"]),
            VfpParam.RATE: VFPPROD_FLO(self._data["RATE_TYPE"]),
        }
        self._param_units = {
            VfpParam.THP: VFPPROD_UNITS[self._unit_type.value]["THP"][
                self.param_types[VfpParam.THP].value
            ],
            VfpParam.WFR: VFPPROD_UNITS[self._unit_type.value]["WFR"][
                self.param_types[VfpParam.WFR].value
            ],
            VfpParam.GFR: VFPPROD_UNITS[self._unit_type.value]["GFR"][
                self.param_types[VfpParam.GFR].value
            ],
            VfpParam.ALQ: VFPPROD_UNITS[self._unit_type.value]["ALQ"][
                self.param_types[VfpParam.ALQ].value
            ],
            VfpParam.RATE: VFPPROD_UNITS[self._unit_type.value]["FLO"][
                self.param_types[VfpParam.RATE].value
            ],
        }
        self._bhp_table = np.array(self._data["BHP_TABLE"])

        # pylint: disable=too-many-function-args
        self._reshaped_bhp_table = self._bhp_table.reshape(
            len(self.params[VfpParam.THP]),
            len(self.params[VfpParam.WFR]),
            len(self.params[VfpParam.GFR]),
            len(self.params[VfpParam.ALQ]),
            len(self.params[VfpParam.RATE]),
        )

    def get_rate_label(self) -> str:
        return f"""
{self.param_types[VfpParam.RATE].value.capitalize()} rate ({self._param_units[VfpParam.RATE]})
"""

    def get_bhp_label(self, pressure_type: PressureType) -> str:
        return f"{pressure_type.value} ({self._param_units[VfpParam.THP]})"

    def get_bhp_series(
        self,
        pressure_type: PressureType,
        thp_idx: int,
        wfr_idx: int,
        gfr_idx: int,
        alq_idx: int,
    ) -> List[float]:
        """Returns a series of bhp values for the given vfp parameter indices.
        The series has the same length as the rate values.

        If pressure_type is DP then the THP at the given thp index is subtracted
        from all the BHP values.
        """
        bhp_values = self._reshaped_bhp_table[thp_idx][wfr_idx][gfr_idx][alq_idx]
        if pressure_type == PressureType.BHP:
            return bhp_values
        if pressure_type == PressureType.DP:
            return bhp_values - self.params[VfpParam.THP][thp_idx]
        raise ValueError(f"PressureType {pressure_type} not implemented")

    def get_values(
        self, vfp_param: VfpParam, indices: Optional[List[int]] = None
    ) -> List[float]:
        """Returns the values for a given vfp param.

        If a list of indices is given, then only the values for those
        indices is returned.
        """
        if indices is None:
            return list(self.params[vfp_param].values())
        return [self.params[vfp_param][idx] for idx in indices]

    def get_metadata_markdown(self) -> str:
        """Returns a markdown with all the table metadata."""
        thp_values = ", ".join([str(val) for val in self.params[VfpParam.THP].values()])
        wfr_values = ", ".join([str(val) for val in self.params[VfpParam.WFR].values()])
        gfr_values = ", ".join([str(val) for val in self.params[VfpParam.GFR].values()])
        alq_values = ", ".join([str(val) for val in self.params[VfpParam.ALQ].values()])
        rate_values = ", ".join(
            [str(val) for val in self.params[VfpParam.RATE].values()]
        )
        return f"""
> - **VFP type**: {self._vfp_type.name}
> - **Table number**: {self._table_number}
> - **Units**: {self._unit_type.name}
> - **Datum**: {self._datum}
> - **THP type**: {self.param_types[VfpParam.THP].name} ({self._param_units[VfpParam.THP]})
> - **WFR type**: {self.param_types[VfpParam.WFR].name} ({self._param_units[VfpParam.WFR]})
> - **GFR type**: {self.param_types[VfpParam.GFR].name} ({self._param_units[VfpParam.GFR]})
> - **ALQ type**: {self.param_types[VfpParam.ALQ].name} ({self._param_units[VfpParam.ALQ]})
> - **Rate type**: {self.param_types[VfpParam.RATE].name} ({self._param_units[VfpParam.RATE]})
> - **THP values**: {thp_values}
> - **WFR values**: {wfr_values}
> - **GFR values**: {gfr_values}
> - **ALQ values**: {alq_values}
> - **Rate values**: {rate_values}
        """


class VfpDataModel:
    """Class for loading a keeping all the VFP tables."""

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        vfp_file_pattern: str,
        ensemble: Optional[str] = None,
        realization: Optional[int] = None,
    ):
        if ensemble is not None:
            if isinstance(ensemble, list):
                raise TypeError(
                    'Incorrent argument type, "ensemble" must be a string instead of a list'
                )

            ens_path = webviz_settings.shared_settings["scratch_ensembles"][ensemble]

            if realization is None:
                raise ValueError('Incorrent arguments, "realization" must be specified')

            ens_path = webviz_settings.shared_settings["scratch_ensembles"][ensemble]
            # replace realization in string from scratch_ensemble with input realization
            ens_path = re.sub(
                "realization-[^/]", f"realization-{realization}", ens_path
            )
            self._vfp_file_pattern = f"{ens_path}/{vfp_file_pattern}"
        else:
            self._vfp_file_pattern = vfp_file_pattern

        self._vfp_files = json.load(_discover_files(self._vfp_file_pattern))
        if not self._vfp_files:
            raise FileNotFoundError(
                "No VFP arrow files found matching input file pattern."
            )

        self._vfp_tables = {}
        for table_name, file_name in self._vfp_files.items():
            try:
                self._vfp_tables[table_name] = VfpTable(file_name)
            except NotImplementedError as exc:
                logging.warning(exc)

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[Dict]]]:
        return [(_discover_files, [{"file_pattern": self._vfp_file_pattern}]),] + [
            (_read_vfp_arrow, [{"filename": filename}])
            for filename in self._vfp_files.values()
        ]

    @property
    def vfp_names(self) -> List[str]:
        """Return unique vfp names"""
        return list(self._vfp_tables.keys())

    def get_vfp_table(self, vfp_name: str) -> VfpTable:
        """Returns a VfpTable object corresponding to the given table number"""
        if not vfp_name in self._vfp_tables:
            raise ValueError(f"Vfp Table: {vfp_name} not found.")
        return self._vfp_tables[vfp_name]


@webvizstore
def _discover_files(file_pattern: str) -> io.BytesIO:
    """Returns all the files that matches the input file pattern."""
    files = {
        file_name.split("/")[-1].replace(".arrow", ""): file_name
        for file_name in glob.glob(file_pattern)
    }
    return io.BytesIO(json.dumps(files).encode())


@webvizstore
def _read_vfp_arrow(filename: str) -> io.BytesIO:
    """Function to read the vfp arrow files and return them as
    a io.BytesIO object in order to be stored as portable.

    Uses the pyarrow2basic_data function from ecl2df in order
    to convert the pyarrow table into a dictionary. But then
    the columns have to be converted to strings, or lists in order
    to be encoded.
    """
    source = pa.memory_map(filename, "r")
    reader = pa.ipc.RecordBatchFileReader(source)
    pa_table = reader.read_all()
    vfp_dict = pyarrow2basic_data(pa_table)

    for key, _ in vfp_dict.items():
        # Convert types to strings
        if key.endswith("_TYPE"):
            vfp_dict[key] = str(vfp_dict[key].value)
        # Convert ndarrays to lists
        if (
            key.endswith("_VALUES")
            or key.endswith("_TABLE")
            or key.endswith("_INDICES")
        ):
            vfp_dict[key] = vfp_dict[key].tolist()

    return io.BytesIO(json.dumps(vfp_dict).encode())



from typing import Any, Dict, List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import ColorAndSizeByType


class MapSettings(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLE = "map-ensemble"
        SIZE_BY = "map-size-by"
        COLOR_BY = "map-color-by"
        DATE_RANGE = "map-date-range"
        ZONES = "map-zones"

    def __init__(
        self,
        ensembles: List[str],
        zones: List[str],
        date_marks: Dict[str, Dict[str, Any]],
        date_range_min: int,
        date_range_max: int,
    ) -> None:
        super().__init__("Map settings")
        self._ensembles = ensembles
        self._zone_names = zones
        self._date_marks = date_marks
        self._date_range_min = date_range_min
        self._date_range_max = date_range_max

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Ensemble",
                id=self.register_component_unique_id(self.Ids.ENSEMBLE),
                options=[{"label": ens, "value": ens} for ens in self._ensembles],
                value=self._ensembles[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Size points by",
                id=self.register_component_unique_id(self.Ids.SIZE_BY),
                options=[
                    {
                        "label": "Standard Deviation",
                        "value": ColorAndSizeByType.STDDEV,
                    },
                    {
                        "label": "Misfit",
                        "value": ColorAndSizeByType.MISFIT,
                    },
                ],
                value=ColorAndSizeByType.MISFIT,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Color points by",
                id=self.register_component_unique_id(self.Ids.COLOR_BY),
                options=[
                    {
                        "label": "Misfit",
                        "value": ColorAndSizeByType.MISFIT,
                    },
                    {
                        "label": "Standard Deviation",
                        "value": ColorAndSizeByType.STDDEV,
                    },
                    {
                        "label": "Year",
                        "value": ColorAndSizeByType.YEAR,
                    },
                ],
                value=ColorAndSizeByType.STDDEV,
                clearable=False,
            ),
            wcc.RangeSlider(
                label="Filter date range",
                id=self.register_component_unique_id(self.Ids.DATE_RANGE),
                min=self._date_range_min,
                max=self._date_range_max,
                value=[
                    self._date_range_min,
                    self._date_range_max,
                ],
                marks=self._date_marks,
            ),
            wcc.Label(
                "Zone filter",
            ),
            wcc.SelectWithLabel(
                size=min(10, len(self._zone_names)),
                id=self.register_component_unique_id(self.Ids.ZONES),
                options=[{"label": name, "value": name} for name in self._zone_names],
                value=self._zone_names,
                multi=True,
            ),
        ]



from typing import Any, Dict, List, Optional, Tuple

import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import Input, Output, State, callback, html
from dash.development.base_component import Component
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC, ViewABC

from ..._types import NodeType, StatOptions, TreeModeOptions
from ..._utils import EnsembleGroupTreeData
from ._view_element import GroupTreeViewElement


class ViewControls(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLE = "ensemble"
        TREEMODE = "tree-mode"

    def __init__(self, ensembles: List[str]) -> None:
        super().__init__("Controls")
        self._ensembles = ensembles

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(self.Ids.ENSEMBLE),
                label="Ensemble",
                clearable=False,
                options=[{"label": ens, "value": ens} for ens in self._ensembles],
                value=self._ensembles[0],
            ),
            wcc.RadioItems(
                id=self.register_component_unique_id(self.Ids.TREEMODE),
                label="Statistics or realization",
            ),
        ]


class ViewOptions(SettingsGroupABC):
    class Ids(StrEnum):
        STATISTICAL_OPTIONS = "statistical-options"
        STATISTICS = "statistics"
        SINGLE_REAL_OPTIONS = "single-real-options"
        REALIZATION = "realization"

    def __init__(self, group_tree_data: Dict[str, EnsembleGroupTreeData]) -> None:
        super().__init__("Options")
        self.group_tree_data = group_tree_data

    def layout(self) -> List[Component]:
        return [
            html.Div(
                id=self.register_component_unique_id(
                    ViewOptions.Ids.STATISTICAL_OPTIONS
                ),
                children=[
                    wcc.RadioItems(
                        id=self.register_component_unique_id(
                            ViewOptions.Ids.STATISTICS
                        ),
                        options=[
                            {"label": "Mean", "value": StatOptions.MEAN},
                            {"label": "P10 (high)", "value": StatOptions.P10},
                            {
                                "label": "P50 (median)",
                                "value": StatOptions.P50,
                            },
                            {"label": "P90 (low)", "value": StatOptions.P90},
                            {"label": "Maximum", "value": StatOptions.MAX},
                            {"label": "Minimum", "value": StatOptions.MIN},
                        ],
                    )
                ],
            ),
            html.Div(
                id=self.register_component_unique_id(
                    ViewOptions.Ids.SINGLE_REAL_OPTIONS
                ),
                children=[
                    wcc.Dropdown(
                        label="Realization",
                        id=self.register_component_unique_id(
                            ViewOptions.Ids.REALIZATION
                        ),
                        options=[],
                        value=None,
                        multi=False,
                    )
                ],
            ),
        ]


class ViewFilters(SettingsGroupABC):
    class Ids(StrEnum):
        TOUR_STEP = "tour-step"
        PROD_INJ_OTHER = "prod-inj-other"

    def __init__(self) -> None:
        super().__init__("Filters")

    def layout(self) -> html.Div:
        return html.Div(
            id=self.register_component_unique_id(self.Ids.TOUR_STEP),
            children=[
                wcc.SelectWithLabel(
                    id=self.register_component_unique_id(self.Ids.PROD_INJ_OTHER),
                    label="Prod/Inj/Other",
                    options=[
                        {"label": "Production", "value": NodeType.PROD},
                        {"label": "Injection", "value": NodeType.INJ},
                        {"label": "Other", "value": NodeType.OTHER},
                    ],
                    value=[
                        NodeType.PROD,
                        NodeType.INJ,
                        NodeType.OTHER,
                    ],
                    multi=True,
                    size=3,
                )
            ],
        )


class GroupTreeView(ViewABC):
    class Ids(StrEnum):
        VIEW_ELEMENT = "view-element"
        CONTROLS = "controls"
        OPTIONS = "options"
        FILTERS = "filters"
        GROUPTREE_COMPONENT = "grouptree-component"

    def __init__(self, group_tree_data: Dict[str, EnsembleGroupTreeData]) -> None:
        super().__init__("Group Tree")
        self._group_tree_data = group_tree_data

        self.add_settings_group(
            ViewControls(list(self._group_tree_data.keys())), GroupTreeView.Ids.CONTROLS
        )
        self.add_settings_group(
            ViewOptions(self._group_tree_data), GroupTreeView.Ids.OPTIONS
        )
        self.add_settings_group(ViewFilters(), GroupTreeView.Ids.FILTERS)

        main_column = self.add_column()
        main_column.add_view_element(
            GroupTreeViewElement(), GroupTreeView.Ids.VIEW_ELEMENT
        )
        self._group_tree_component_id = self.view_element(
            GroupTreeView.Ids.VIEW_ELEMENT
        ).register_component_unique_id(GroupTreeView.Ids.GROUPTREE_COMPONENT)

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.TREEMODE)
                .to_string(),
                "options",
            ),
            Output(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.TREEMODE)
                .to_string(),
                "value",
            ),
            Output(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.STATISTICS)
                .to_string(),
                "value",
            ),
            Output(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.REALIZATION)
                .to_string(),
                "options",
            ),
            Output(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.TREEMODE)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.STATISTICS)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _update_ensemble_options(
            ensemble_name: str,
            tree_mode_state: Optional[TreeModeOptions],
            stat_option_state: Optional[StatOptions],
            real_state: Optional[int],
        ) -> Tuple[List[Dict[str, Any]], str, str, List[Dict[str, Any]], Optional[int]]:
            """Updates the selection options when the ensemble value changes"""
            tree_mode_options: List[Dict[str, Any]] = [
                {
                    "label": "Statistics",
                    "value": TreeModeOptions.STATISTICS,
                },
                {
                    "label": "Single realization",
                    "value": TreeModeOptions.SINGLE_REAL,
                },
            ]
            tree_mode = (
                tree_mode_state
                if tree_mode_state is not None
                else TreeModeOptions.STATISTICS
            )
            stat_option = (
                stat_option_state if stat_option_state is not None else StatOptions.MEAN
            )

            ensemble = self._group_tree_data[ensemble_name]
            if not ensemble.tree_is_equivalent_in_all_real():
                tree_mode_options[0]["label"] = "Ensemble mean (disabled)"
                tree_mode_options[0]["disabled"] = True
                tree_mode = TreeModeOptions.SINGLE_REAL

            unique_real = ensemble.get_unique_real()

            return (
                tree_mode_options,
                tree_mode,
                stat_option,
                [{"label": real, "value": real} for real in unique_real],
                real_state if real_state in unique_real else min(unique_real),
            )

        @callback(
            Output(
                self.view_element(GroupTreeView.Ids.VIEW_ELEMENT)
                .component_unique_id(GroupTreeViewElement.Ids.COMPONENT)
                .to_string(),
                "children",
            ),
            Input(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.TREEMODE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.STATISTICS)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.REALIZATION)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(GroupTreeView.Ids.FILTERS)
                .component_unique_id(ViewFilters.Ids.PROD_INJ_OTHER)
                .to_string(),
                "value",
            ),
            State(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _render_grouptree(
            tree_mode: TreeModeOptions,
            stat_option: StatOptions,
            real: int,
            node_types: List[NodeType],
            ensemble_name: str,
        ) -> list:
            """This callback updates the input dataset to the Grouptree component."""
            data, edge_options, node_options = self._group_tree_data[
                ensemble_name
            ].create_grouptree_dataset(
                tree_mode,
                stat_option,
                real,
                node_types,
            )

            return [
                wsc.GroupTree(
                    id=self._group_tree_component_id,
                    data=data,
                    edge_options=edge_options,
                    node_options=node_options,
                ),
            ]

        @callback(
            Output(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.STATISTICAL_OPTIONS)
                .to_string(),
                "style",
            ),
            Output(
                self.settings_group(GroupTreeView.Ids.OPTIONS)
                .component_unique_id(ViewOptions.Ids.SINGLE_REAL_OPTIONS)
                .to_string(),
                "style",
            ),
            Input(
                self.settings_group(GroupTreeView.Ids.CONTROLS)
                .component_unique_id(ViewControls.Ids.TREEMODE)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        def _show_hide_single_real_options(
            tree_mode: TreeModeOptions,
        ) -> Tuple[Dict[str, str], Dict[str, str]]:
            if tree_mode is TreeModeOptions.STATISTICS:
                return {"display": "block"}, {"display": "none"}
            return {"display": "none"}, {"display": "block"}



from typing import Any, Callable, Dict, List, Optional

import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import dash_table, html

from ..models import PropertyStatisticsModel
from .selector_view import (
    delta_ensemble_selector,
    ensemble_selector,
    filter_selector,
    property_selector,
    source_selector,
)


def surface_select_view(get_uuid: Callable, tab: str) -> html.Div:
    return html.Div(
        id=get_uuid("surface-select"),
        style={"width": "75%"},
        children=wcc.Dropdown(
            label="Surface statistics",
            id={"id": get_uuid("surface-type"), "tab": tab},
            options=[
                {"label": "Mean", "value": "mean"},
                {"label": "Standard Deviation", "value": "stddev"},
                {"label": "Minimum", "value": "min"},
                {"label": "Maximum", "value": "max"},
                {"label": "P10", "value": "p10"},
                {"label": "P90", "value": "p90"},
            ],
            clearable=False,
            value="mean",
        ),
    )


def surface_view(
    get_uuid: Callable, ensemble: str, layers: list, synced_ids: list = None
) -> wsc.LeafletMap:
    return html.Div(
        style={"height": "22vh"},
        children=wsc.LeafletMap(
            id=f"{get_uuid('surface-view-delta')}{ensemble}",
            layers=layers,
            syncedMaps=[
                f"{get_uuid('surface-view-delta')}{s_id}" for s_id in synced_ids
            ]
            if synced_ids is not None
            else [],
            unitScale={},
            autoScaleMap=True,
            minZoom=-19,
            updateMode="replace",
            mouseCoords={"position": "bottomright"},
            colorBar={"position": "bottomleft"},
        ),
    )


# pylint: disable=too-many-arguments
def surface_views(
    get_uuid: Callable,
    ensemble: str,
    delta_ensemble: str,
    ens_layer: dict,
    delta_ens_layer: dict,
    diff_layer: dict,
    prop: str,
    zone: str,
    statistic: str = "Mean",
) -> html.Div:
    return html.Div(
        style={"height": "76vh"},
        children=[
            wcc.Label(
                children=f"{statistic} for {prop}, {zone} in {ensemble}",
            ),
            surface_view(
                get_uuid=get_uuid,
                ensemble=ensemble,
                synced_ids=[delta_ensemble, "diff"],
                layers=[ens_layer],
            ),
            wcc.Label(
                children=f"{statistic} for {prop}, {zone} in {delta_ensemble}",
            ),
            surface_view(
                get_uuid=get_uuid,
                ensemble=delta_ensemble,
                synced_ids=[ensemble, "diff"],
                layers=[delta_ens_layer],
            ),
            wcc.Label(
                children=f"{statistic} for {prop}, {zone} in {ensemble} - {delta_ensemble}",
            ),
            surface_view(
                get_uuid=get_uuid,
                ensemble="diff",
                synced_ids=[delta_ensemble, ensemble],
                layers=[diff_layer],
            ),
        ],
    )


def selector_view(
    get_uuid: Callable, property_model: PropertyStatisticsModel
) -> html.Div:
    return wcc.Frame(
        style={"height": "80vh", "overflowY": "auto"},
        children=[
            wcc.Selectors(
                label="Selectors",
                children=[
                    ensemble_selector(
                        get_uuid=get_uuid,
                        ensembles=property_model.ensembles,
                        tab="delta",
                    ),
                    delta_ensemble_selector(
                        get_uuid=get_uuid,
                        ensembles=property_model.ensembles,
                        tab="delta",
                    ),
                    property_selector(
                        get_uuid=get_uuid,
                        properties=property_model.properties,
                        tab="delta",
                    ),
                    source_selector(
                        get_uuid=get_uuid, sources=property_model.sources, tab="delta"
                    ),
                ],
            ),
            wcc.Selectors(
                label="Filters",
                children=[
                    filter_selector(
                        get_uuid=get_uuid, property_model=property_model, tab="delta"
                    )
                ],
            ),
        ],
    )


def delta_avg_view() -> html.Div:
    return wcc.FlexColumn(
        children=[],
    )


def property_delta_view(
    get_uuid: Callable,
    property_model: PropertyStatisticsModel,
    surface_folders: Optional[Dict] = None,
) -> wcc.FlexBox:
    table_surf_options = [{"label": "Table view", "value": "table"}]
    if surface_folders is not None:
        table_surf_options.append(
            {"label": "Surface view (click on bar in chart)", "value": "surface"}
        )
    return wcc.FlexBox(
        style={"margin": "20px"},
        children=[
            wcc.FlexColumn(
                children=selector_view(get_uuid=get_uuid, property_model=property_model)
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"height": "80vh"},
                    children=[
                        wcc.RadioItems(
                            vertical=False,
                            id=get_uuid("delta-sort"),
                            options=[
                                {"label": "Sort by Average", "value": "Avg"},
                                {
                                    "label": "Sort by Standard Deviation",
                                    "value": "Stddev",
                                },
                            ],
                            value="Avg",
                        ),
                        wcc.Graph(
                            id=get_uuid("delta-bar-graph"),
                            config={"displayModeBar": False},
                            style={"height": "75vh"},
                        ),
                    ],
                ),
            ),
            wcc.FlexColumn(
                flex=4,
                children=wcc.Frame(
                    style={"height": "80vh"},
                    color="white",
                    highlight=False,
                    children=[
                        wcc.RadioItems(
                            id=get_uuid("delta-switch-table-surface"),
                            vertical=False,
                            options=table_surf_options,
                            value="table",
                        ),
                        html.Div(id=get_uuid("delta-table-surface-wrapper")),
                    ],
                ),
            ),
        ],
    )


def table_view(data: List[Any], columns: List[Any]) -> html.Div:
    return dash_table.DataTable(
        sort_action="native",
        page_action="native",
        filter_action="native",
        style_table={"height": "74vh", "overflow": "auto", "fontSize": 13},
        data=data,
        columns=columns,
        style_cell={"textAlign": "center"},
        style_cell_conditional=[{"if": {"column_id": "label|"}, "textAlign": "left"}],
        merge_duplicate_headers=True,
    )



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import ColorAndSizeByType


class SizeColorSettings(SettingsGroupABC):
    class Ids(StrEnum):
        CROSSPLOT_COLOR_BY = "crossplot-color-by"
        CROSSPLOT_SIZE_BY = "crossplot-size-by"

    def __init__(self) -> None:
        super().__init__("Crossplot options")

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Color by",
                id=self.register_component_unique_id(self.Ids.CROSSPLOT_COLOR_BY),
                options=[
                    {
                        "label": "Misfit",
                        "value": ColorAndSizeByType.MISFIT,
                    },
                    {
                        "label": "Standard Deviation",
                        "value": ColorAndSizeByType.STDDEV,
                    },
                ],
                value=ColorAndSizeByType.STDDEV,
                clearable=False,
            ),
            wcc.Dropdown(
                label="Size by",
                id=self.register_component_unique_id(self.Ids.CROSSPLOT_SIZE_BY),
                options=[
                    {
                        "label": "Standard Deviation",
                        "value": ColorAndSizeByType.STDDEV,
                    },
                    {
                        "label": "Misfit",
                        "value": ColorAndSizeByType.MISFIT,
                    },
                ],
                value=ColorAndSizeByType.MISFIT,
                clearable=False,
            ),
        ]



import plotly.express as px
import plotly.graph_objs as go


def color_figure(
    colors: list,
    bargap: float = None,
    height: float = None,
):
    """
    Create bar chart with colors, can e.g. be used as a color selector
    by retrieving clickdata.
    The color argument is a list of items where the individual items
    are either a name of a built-in plotly colorscale, or a list of colors.
    """
    color_lists = []
    for item in colors:
        if isinstance(item, list) and item:
            color_lists.append(item)
        if isinstance(item, str):
            color_lists.append(get_px_colors(item))

    return go.Figure(
        data=[
            go.Bar(
                orientation="h",
                y=[str(idx)] * len(clist),
                x=[1] * len(clist),
                customdata=list(range(len(clist))),
                marker=dict(color=clist),
                hovertemplate="%{marker.color}<extra></extra>",
            )
            for idx, clist in enumerate(reversed(color_lists))
        ],
        layout=dict(
            title=None,
            barmode="stack",
            barnorm="fraction",
            bargap=bargap if bargap is not None else 0.5,
            showlegend=False,
            xaxis=dict(range=[-0.02, 1.02], showticklabels=False, showgrid=False),
            yaxis_showticklabels=False,
            height=height if height is not None else 20 * len(color_lists),
            margin=dict(l=0, r=0, t=0, b=0),
            paper_bgcolor="rgba(0,0,0,0)",
            plot_bgcolor="rgba(0,0,0,0)",
        ),
    )


def get_px_colors(px_cscale: str):
    for cmodule in [
        px.colors.diverging,
        px.colors.sequential,
        px.colors.cyclical,
        px.colors.qualitative,
    ]:
        if px_cscale in cmodule.__dict__:
            return cmodule.__dict__[px_cscale]
    return []



from typing import Callable, Tuple, Union

import pandas as pd
import plotly.graph_objects as go
from dash import Dash, Input, Output, State, callback_context, no_update
from dash.exceptions import PreventUpdate

from webviz_subsurface._figures import BarChart, ScatterPlot, TimeSeriesFigure
from webviz_subsurface._utils.colors import hex_to_rgba_str, rgba_to_hex
from webviz_subsurface._utils.dataframe_utils import (
    correlate_response_with_dataframe,
    merge_dataframes_on_realization,
)

from .._utils import datetime_utils
from ..models import (
    ParametersModel,
    ProviderTimeSeriesDataModel,
    SimulationTimeSeriesModel,
)


# pylint: disable=too-many-statements
def parameter_response_controller(
    app: Dash,
    get_uuid: Callable,
    vectormodel: Union[SimulationTimeSeriesModel, ProviderTimeSeriesDataModel],
    parametermodel: ParametersModel,
):
    @app.callback(
        Output(get_uuid("vector-vs-time-graph"), "figure"),
        Output(get_uuid("vector-vs-param-scatter"), "figure"),
        Output(get_uuid("vector-corr-graph"), "figure"),
        Output(get_uuid("param-corr-graph"), "figure"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "response"}, "value"),
        Input(get_uuid("vector-selector"), "selectedNodes"),
        Input({"id": get_uuid("parameter-select"), "tab": "response"}, "value"),
        Input(get_uuid("date-selected"), "data"),
        Input(get_uuid("vector-filter-store"), "data"),
        Input(get_uuid("visualization"), "value"),
        Input({"id": get_uuid("plot-options"), "tab": "response"}, "data"),
        Input({"id": get_uuid("parameter-filter"), "type": "data-store"}, "data"),
        State(get_uuid("param-corr-graph"), "figure"),
        State(get_uuid("vector-corr-graph"), "figure"),
    )
    # pylint: disable=too-many-locals, too-many-arguments
    def _update_graphs(
        ensemble: str,
        vector: str,
        parameter: Union[None, dict],
        date: str,
        column_keys: str,
        visualization: str,
        options: str,
        real_filter: dict,
        corr_p_fig: dict,
        corr_v_fig: dict,
    ) -> Tuple[dict, dict, dict, dict]:
        """
        Main callback to update plots. Initially all plots are generated,
        while only relevant plots are updated in subsequent callbacks
        """
        ctx = callback_context.triggered[0]["prop_id"].split(".")[0]
        if not ctx or not vector:
            raise PreventUpdate

        vector = vector[0]
        date = datetime_utils.from_str(date)
        realizations = real_filter[ensemble]
        color = options["color"] if options["color"] is not None else "#007079"

        if len(realizations) <= 1:
            return [empty_figure()] * 4

        filtered_vectors = (
            vectormodel.filter_vectors(column_keys) if column_keys is not None else []
        )
        vectors = list(set(filtered_vectors + [vector]))

        # Get dataframe with vectors and dataframe with parameters and merge
        vector_df = vectormodel.get_vector_df(
            ensemble=ensemble, realizations=realizations, vectors=vectors
        )
        if date not in vector_df["DATE"].values:
            return [empty_figure("Selected date does not exist for ensemble")] * 4
        if vector not in vector_df:
            return [empty_figure("Selected vector does not exist for ensemble")] * 4

        param_df = parametermodel.get_parameter_df_for_ensemble(ensemble, realizations)
        merged_df = merge_dataframes_on_realization(
            dframe1=vector_df[vector_df["DATE"] == date], dframe2=param_df
        )
        # Make correlation figure for vector
        if options["autocompute_corr"]:
            corr_v_fig = make_correlation_figure(
                merged_df, response=vector, corrwith=parametermodel.parameters
            ).figure

        # Get clicked parameter correlation bar or largest bar initially
        parameter = (
            parameter if parameter is not None else corr_v_fig["data"][0]["y"][-1]
        )
        corr_v_fig = color_corr_bars(corr_v_fig, parameter, color, options["opacity"])

        if not filtered_vectors:
            text = (
                "Select vectors for parameter correlation to correlate"
                if not bool(column_keys)
                else "No vectors match selected filter"
            )
            corr_p_fig = empty_figure(text)
        else:
            # Make correlation figure for parameter
            if options["autocompute_corr"]:
                corr_p_fig = make_correlation_figure(
                    merged_df, response=parameter, corrwith=vectors
                ).figure

            corr_p_fig = color_corr_bars(corr_p_fig, vector, color, options["opacity"])

        # Create scatter plot of vector vs parameter
        scatter_fig = ScatterPlot(
            merged_df,
            response=vector,
            param=parameter,
            color=color,
            title=f"{vector} vs {parameter}",
            plot_trendline=True,
        )
        scatter_fig.update_color(color, options["opacity"])

        # Make timeseries graph
        df_value_norm = parametermodel.get_real_and_value_df(
            ensemble, parameter=parameter, normalize=True
        )
        timeseries_fig = TimeSeriesFigure(
            dframe=merge_dataframes_on_realization(
                vector_df[["DATE", "REAL", vector]], df_value_norm
            ),
            visualization=visualization,
            vector=vector,
            ensemble=ensemble,
            dateline=date if options["show_dateline"] else None,
            historical_vector_df=vectormodel.get_historical_vector_df(vector, ensemble),
            color_col=parameter,
            line_shape_fallback=vectormodel.line_shape_fallback,
        ).figure

        return timeseries_fig, scatter_fig.figure, corr_v_fig, corr_p_fig

    @app.callback(
        Output(get_uuid("date-slider"), "value"),
        Input(get_uuid("vector-vs-time-graph"), "clickData"),
        State(get_uuid("date-selected"), "data"),
    )
    def _update_date_from_clickdata(timeseries_clickdata: Union[None, dict], date):
        """Update date-slider from clickdata"""
        date = (
            timeseries_clickdata.get("points", [{}])[0]["x"]
            if timeseries_clickdata is not None
            else date
        )
        return vectormodel.dates.index(datetime_utils.from_str(date))

    @app.callback(
        Output(get_uuid("vector-filter-store"), "data"),
        Output(get_uuid("submit-vector-filter"), "style"),
        Input(get_uuid("submit-vector-filter"), "n_clicks"),
        Input(get_uuid("vector-filter"), "value"),
        State(get_uuid("vector-filter-store"), "data"),
        State(get_uuid("submit-vector-filter"), "style"),
    )
    def _update_vector_filter_store_and_button_style(
        _n_click: int, vector_filter: str, stored: str, style: dict
    ):
        """Update vector-filter-store if submit button is clicked and
        style of submit button"""
        ctx = callback_context.triggered[0]["prop_id"]
        button_click = "submit" in ctx
        insync = stored == vector_filter
        style["background-color"] = "#E8E8E8" if insync or button_click else "#7393B3"
        style["color"] = "#555" if insync or button_click else "#fff"
        return vector_filter if button_click else no_update, style

    @app.callback(
        Output(get_uuid("date-selected"), "data"),
        Input(get_uuid("date-slider"), "value"),
    )
    def _update_date(dateidx: int):
        """Update selected date from date-slider"""
        return datetime_utils.to_str(vectormodel.dates[dateidx])

    @app.callback(
        Output(get_uuid("date-selected-text"), "children"),
        Input(get_uuid("date-slider"), "drag_value"),
        prevent_initial_call=True,
    )
    def _update_date(dateidx: int):
        """Update selected date text on date-slider drag"""
        return datetime_utils.to_str(vectormodel.dates[dateidx])

    @app.callback(
        Output({"id": get_uuid("plot-options"), "tab": "response"}, "data"),
        Input({"id": get_uuid("checkbox-options"), "tab": "response"}, "value"),
        Input({"id": get_uuid("color-selector"), "tab": "response"}, "clickData"),
        Input({"id": get_uuid("opacity-selector"), "tab": "response"}, "value"),
        State({"id": get_uuid("plot-options"), "tab": "response"}, "data"),
    )
    def _update_plot_options(
        checkbox_options: list,
        color_clickdata: str,
        opacity: float,
        plot_options: dict,
    ):
        """Combine plot options in one dictionary"""
        ctx = callback_context.triggered[0]["prop_id"].split(".")[0]
        if plot_options is not None and not ctx:
            raise PreventUpdate
        if color_clickdata is not None:
            color = color_clickdata["points"][0]["marker.color"]
            if "rgb" in color:
                color = rgba_to_hex(color)

        return dict(
            show_dateline="DateLine" in checkbox_options,
            autocompute_corr="AutoCompute" in checkbox_options,
            color=None if color_clickdata is None else color,
            opacity=opacity,
            ctx=ctx,
        )

    @app.callback(
        Output(get_uuid("vector-selector"), "selectedTags"),
        Input(get_uuid("param-corr-graph"), "clickData"),
    )
    def _update_vectorlist(corr_param_clickdata: dict):
        """Update the selected vector value from clickdata"""
        if corr_param_clickdata is None:
            raise PreventUpdate
        vector_selected = corr_param_clickdata.get("points", [{}])[0].get("y")
        return [vector_selected]

    @app.callback(
        Output({"id": get_uuid("parameter-select"), "tab": "response"}, "options"),
        Output({"id": get_uuid("parameter-select"), "tab": "response"}, "value"),
        Input(get_uuid("vector-corr-graph"), "clickData"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "response"}, "value"),
        State({"id": get_uuid("parameter-select"), "tab": "response"}, "value"),
    )
    def _update_parameter_selected(
        corr_vector_clickdata: Union[None, dict], ensemble: str, selected_parameter: str
    ) -> tuple:
        """Update the selected parameter from clickdata, or when ensemble is changed"""
        ctx = callback_context.triggered[0]["prop_id"]
        if ctx == ".":
            raise PreventUpdate
        parameters = parametermodel.pmodel.parameters_per_ensemble[ensemble]
        options = [{"label": i, "value": i} for i in parameters]
        if "vector-corr-graph" in ctx:
            return options, corr_vector_clickdata.get("points", [{}])[0].get("y")
        return options, selected_parameter if selected_parameter in parameters else None

    @app.callback(
        Output({"id": get_uuid("parameter-filter"), "type": "ensemble-update"}, "data"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "response"}, "value"),
    )
    def _update_parameter_filter_selection(ensemble: str):
        """Update ensemble in parameter filter"""
        return [ensemble]

    @app.callback(
        Output(get_uuid("param-filter-wrapper"), "style"),
        Input(get_uuid("display-paramfilter"), "n_clicks"),
        State(get_uuid("param-filter-wrapper"), "style"),
    )
    def _show_hide_parameter_filter(_n_click: list, style: dict):
        """Display/hide parameter filter"""
        if _n_click is None:
            raise PreventUpdate
        style.update(display="block" if style["display"] == "none" else "none")
        return style

    @app.callback(
        Output(get_uuid("options-dialog"), "open"),
        Input(get_uuid("options-button"), "n_clicks"),
        State(get_uuid("options-dialog"), "open"),
    )
    def open_close_options_dialog(_n_click: list, is_open: bool) -> bool:
        if _n_click is not None:
            return not is_open
        raise PreventUpdate


def make_correlation_figure(df: pd.DataFrame, response: str, corrwith: list):
    """Create a bar plot with correlations for chosen response"""
    corrseries = correlate_response_with_dataframe(df, response, corrwith)
    return BarChart(
        corrseries, n_rows=15, title=f"Correlations with {response}", orientation="h"
    )


def color_corr_bars(
    figure: dict,
    selected_bar: str,
    color: str,
    opacity: float,
    color_selected="#FF1243",
):
    """
    Set colors to the correlation plot bar,
    with separate color for the selected bar
    """
    if "data" in figure:
        figure["data"][0]["marker"] = {
            "color": [
                hex_to_rgba_str(color, opacity)
                if _bar != selected_bar
                else hex_to_rgba_str(color_selected, 0.8)
                for _bar in figure["data"][0]["y"]
            ],
            "line": {
                "color": [
                    color if _bar != selected_bar else color_selected
                    for _bar in figure["data"][0]["y"]
                ],
                "width": 1.2,
            },
        }
    return figure


def empty_figure(text="No data available for figure") -> go.Figure:
    return go.Figure(
        layout={
            "xaxis": {"visible": False},
            "yaxis": {"visible": False},
            "plot_bgcolor": "white",
            "annotations": [
                {
                    "text": text,
                    "xref": "paper",
                    "yref": "paper",
                    "showarrow": False,
                    "font": {"size": 16},
                }
            ],
        }
    )



from enum import Enum
from typing import Any, Dict, List, Optional

import numpy as np
import xtgeo
from webviz_config.common_cache import CACHE

from webviz_subsurface._models import SurfaceSetModel
from webviz_subsurface._utils.colors import hex_to_rgba_str

from ...._utils.fanchart_plotting import (
    FanchartData,
    LowHighData,
    MinMaxData,
    get_fanchart_traces,
)


class FanChartStatistics(str, Enum):
    MINIMUM = "Min"
    MAXIMUM = "Max"
    P10 = "P10"
    P90 = "P90"


# pylint: disable=too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_plotly_trace_statistical_surface(
    surfaceset: SurfaceSetModel,
    fence_spec: np.ndarray,
    calculation: str,
    name: str,
    legendname: str,
    attribute: str,
    realizations: Optional[List[int]],
    color: str = "red",
    showlegend: bool = False,
    sampling: Optional[str] = "billinear",
) -> Dict[str, Any]:
    """Returns x,y values along a fence for a calculated surface"""
    surface = surfaceset.calculate_statistical_surface(
        name=name,
        attribute=attribute,
        calculation=calculation,
        realizations=realizations,
    )
    fencexy = get_surface_randomline(surface, fence_spec, sampling)
    line_style: Dict[str, Any] = {"width": 4}
    if calculation in ["Min", "Max"]:
        line_style.update({"dash": "dash"})
    return {
        "x": fencexy[:, 0],
        "y": fencexy[:, 1],
        "name": legendname,
        "legendgroup": name,
        "text": f"{legendname} {calculation}",
        "showlegend": showlegend,
        "hoverinfo": "y+x+text",
        "mode": "lines",
        "marker": {"color": hex_to_rgba_str(hex_string=color)},
        "line": line_style,
    }


# pylint: disable=too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_plotly_traces_uncertainty_envelope(
    surfaceset: SurfaceSetModel,
    fence_spec: np.ndarray,
    name: str,
    legendname: str,
    attribute: str,
    realizations: Optional[List[int]],
    color: str = "red",
    showlegend: bool = False,
    sampling: Optional[str] = "billinear",
) -> List:
    """Returns a set of plotly traces representing an uncertainty envelope
    for a surface"""
    values_for_fanchart: Dict[str, np.ma.MaskedArray] = {}
    fan_chart_traces: List = []
    for calculation in FanChartStatistics:
        values = surfaceset.calculate_statistical_surface(
            name=name,
            attribute=attribute,
            calculation=calculation,
            realizations=realizations,
        ).get_randomline(fence_spec, sampling=sampling)
        # Convert to masked array
        values = np.ma.masked_array(values, mask=np.isnan(values))

        if calculation == FanChartStatistics.MINIMUM:
            values_for_fanchart["x"] = values[:, 0]
        values_for_fanchart[FanChartStatistics(calculation)] = values[:, 1]

    # Fanchart plotting requires continuous data series.
    # 1. Create a slice for each non-masked section of y(depth) values.
    #    As the mask is the same for all statistical surfaces,
    #    the minimum surface is randomly used.
    # 2. Make a set of fanchart traces for each slice.

    for unmasked_slice in np.ma.clump_unmasked(
        values_for_fanchart[FanChartStatistics.MINIMUM]
    ):
        fan_chart_data = FanchartData(
            samples=values_for_fanchart["x"][unmasked_slice],
            low_high=LowHighData(
                low_data=values_for_fanchart[FanChartStatistics.P10][unmasked_slice],
                low_name=FanChartStatistics.P10,
                high_data=values_for_fanchart[FanChartStatistics.P90][unmasked_slice],
                high_name=FanChartStatistics.P90,
            ),
            minimum_maximum=MinMaxData(
                minimum=values_for_fanchart[FanChartStatistics.MINIMUM][unmasked_slice],
                maximum=values_for_fanchart[FanChartStatistics.MAXIMUM][unmasked_slice],
            ),
        )
        fan_chart_traces.extend(
            get_fanchart_traces(
                data=fan_chart_data,
                hex_color=color,
                legend_group=name,
                legend_name=legendname,
                show_legend=showlegend,
                show_hoverinfo=True,
            )
        )
    return fan_chart_traces


# pylint: disable=too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_plotly_trace_realization_surface(
    surfaceset: SurfaceSetModel,
    fence_spec: np.ndarray,
    name: str,
    legendname: str,
    attribute: str,
    realization: int,
    showlegend: bool = False,
    sampling: Optional[str] = "billinear",
    color: str = "red",
) -> Dict[str, Any]:
    """Returns a plotly line trace for a surface"""
    surface = surfaceset.get_realization_surface(
        name=name, attribute=attribute, realization=realization
    )
    fencexy = get_surface_randomline(surface, fence_spec, sampling)
    return {
        "x": fencexy[:, 0],
        "y": fencexy[:, 1],
        "name": legendname,
        "legendgroup": name,
        "text": f"{legendname} realization: {realization}",
        "showlegend": showlegend,
        "hoverinfo": "y+x+text",
        "mode": "lines",
        "marker": {"color": hex_to_rgba_str(hex_string=color, opacity=0.5)},
    }


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_plotly_zonelog_trace(
    well: xtgeo.Well,
    zonelog: str,
) -> List[Dict]:
    """Zonetops are extracted from a zonelog and plotted as markers"""

    df = well.dataframe

    # Find zone transitions
    df["transitions"] = df[zonelog].diff()
    logrecord = well.get_logrecord(zonelog)
    traces = []
    for idx in range(0, df.shape[0] - 1):
        # Use current sample if moving upwards in stratigraphy
        # Use next sample if moving downwards in stratigraphy
        if df.iloc[idx + 1]["transitions"] < 0 or df.iloc[idx]["transitions"] > 0:
            traces.append(
                {
                    "x": [df.iloc[idx]["R_HLEN"]],
                    "y": [df.iloc[idx]["Z_TVDSS"]],
                    "marker": {"size": 10, "color": "red"},
                    "showlegend": False,
                    "hoverinfo": "y+name+text",
                    "text": "TVDSS",
                    "hoverlabel": {"namelength": -1},
                    "name": f"Zonetop: <br>{logrecord[df.iloc[idx][zonelog]]}",
                }
            )
    return traces


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_well_xyarray(well: xtgeo.Well) -> List:
    """Returns a copy of the x,y values representing the well fence"""
    dfr = well.dataframe
    zvals = dfr["Z_TVDSS"].values.copy()
    hvals = dfr["R_HLEN"].values.copy()
    return [hvals, zvals]


def get_plotly_trace_well_trajectory(well: xtgeo.Well) -> Dict[str, Any]:
    """Plot the trajectory as a black line"""
    xyarray = get_well_xyarray(well=well)

    return {
        "x": xyarray[0],
        "y": xyarray[1],
        "name": well.name,
        "hoverinfo": "name+text+y",
        "hoverlabel": {"namelength": -1},
        "text": "TVDSS",
        "marker": {"color": "black"},
    }


def get_surface_randomline(
    surface: xtgeo.RegularSurface,
    fence_spec: np.ndarray,
    sampling: Optional[str] = "billinear",
) -> np.ndarray:
    try:
        return surface.get_randomline(fence_spec, sampling=sampling)
    except TypeError:
        # Attempt to handle vertical wells
        fence_spec = np.insert(fence_spec, 0, fence_spec[0] - 10, axis=0)
        fence_spec = np.append(fence_spec, [fence_spec[-1] + 10], axis=0)
        return surface.get_randomline(fence_spec, sampling=sampling)



import itertools
from typing import Any, Dict

import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from webviz_config import WebvizConfigTheme

from ......_utils.colors import StandardColors
from ...._types import NodeType, PressurePlotMode


def create_well_control_figure(
    node_info: Dict[str, Any],
    smry: pd.DataFrame,
    pressure_plot_mode: PressurePlotMode,
    real: int,
    display_ctrlmode_bar: bool,
    shared_xaxes: bool,
    include_bhp: bool,
    theme: WebvizConfigTheme,
) -> Dict:
    # pylint: disable=too-many-locals
    """Creates the plotly figure consisting of multiple subplots, possibly
    sharing the same x-axis.
    """
    node_name = node_info["name"]
    titles = [
        "Number of realizations on different control modes",
        "Network Pressures",
    ]
    rows = 2
    row_heights = [0.5, 0.5]
    if display_ctrlmode_bar and pressure_plot_mode == PressurePlotMode.SINGLE_REAL:
        rows = 3
        display_ctrlmode_bar = True
        row_heights = [0.46, 0.46, 0.08]
        titles.append("Single realization control modes")

    fig = make_subplots(
        rows=rows,
        cols=1,
        shared_xaxes=shared_xaxes,
        vertical_spacing=0.1,
        subplot_titles=titles if titles else ["No vector selected"],
        row_heights=row_heights,
    )
    theme_colors = theme.plotly_theme["layout"]["colorway"]

    # Prepare data
    ctrlmode_sumvec = node_info["ctrlmode_sumvec"]
    smry_ctrlmodes = smry[["DATE", "REAL", ctrlmode_sumvec]].copy()
    # Truncate at -1
    smry_ctrlmodes[ctrlmode_sumvec].clip(-1, None, inplace=True)
    # Replace interpolated values
    smry_ctrlmodes[ctrlmode_sumvec] = smry_ctrlmodes[ctrlmode_sumvec].apply(
        lambda x: "Interpolated" if x % 1 else x
    )

    # Add traces
    add_ctrl_mode_traces(fig, node_info, smry_ctrlmodes)
    add_network_pressure_traces(
        fig, node_info, smry, pressure_plot_mode, real, include_bhp, theme_colors
    )

    if display_ctrlmode_bar and pressure_plot_mode == PressurePlotMode.SINGLE_REAL:
        add_ctrlmode_bar(
            fig,
            node_info,
            smry_ctrlmodes,
            real,
            3,
        )
        fig.update_yaxes(range=[0, 1], row=3, col=1, visible=False)

    fig.update_layout(
        plot_bgcolor="white",
        title=f"Network Analysis for node: {node_name}",
    )
    fig.update_yaxes(
        title="# realizations", range=[0, smry.REAL.nunique()], row=1, col=1
    )
    fig.update_yaxes(autorange="reversed", title="Pressure", row=2, col=1)
    return fig.to_dict()


def add_ctrl_mode_traces(fig: go.Figure, node_info: dict, smry: pd.DataFrame) -> None:
    """Adding area chart traces to the control modes subplot."""
    sumvec = node_info["ctrlmode_sumvec"]
    df = smry.groupby("DATE")[sumvec].value_counts().unstack().fillna(0).reset_index()
    df["Other"] = 0
    categories = get_ctrlmode_categories(node_info["type"])

    for col in [col for col in df.columns if not col in ["DATE", "Other"]]:
        if str(col) in categories:
            name = categories[str(col)]["name"]
            color = categories[str(col)]["color"]
            add_area_trace(fig, df.DATE, df[col], name, color, row=1)
        else:
            df.Other = df.Other + df[col]

    if df.Other.sum() > 0:
        add_area_trace(
            fig,
            df.DATE,
            df.Other,
            categories["Other"]["name"],
            categories["Other"]["color"],
            row=1,
        )


def add_network_pressure_traces(
    fig: go.Figure,
    node_info: Dict[str, Any],
    smry: pd.DataFrame,
    pressure_plot_mode: PressurePlotMode,
    real: int,
    include_bhp: bool,
    theme_colors: list,
) -> None:
    """Adding line traces to the network pressures subplot.

    Missing summary vectors is currently just ignored, but it is possible
    to show a warning or even raise an exception.
    """

    color_iterator = itertools.cycle(theme_colors)
    traces = {}
    for node_network in node_info["networks"]:
        df = get_filtered_smry(
            node_network, node_info["ctrlmode_sumvec"], pressure_plot_mode, real, smry
        )

        for nodedict in node_network["nodes"]:
            if nodedict["type"] == NodeType.WELL_BH and not include_bhp:
                continue
            sumvec = nodedict["pressure"]
            label = nodedict["label"]

            # It the summary vector is not found, do nothing.
            if sumvec in df.columns:
                if label not in traces:
                    traces[label] = {
                        "type": "scatter",
                        "x": [],
                        "y": [],
                        "line": dict(color=next(color_iterator)),
                        "name": label,
                        "showlegend": True,
                        "hovertext": (f"{label}"),
                        "legendgroup": "Nodes",
                        "mode": "lines",
                    }
                traces[label]["x"].extend(list(df["DATE"]))
                traces[label]["y"].extend(list(df[sumvec]))

    for trace in traces.values():
        fig.add_trace(trace, row=2, col=1)


def add_ctrlmode_bar(
    fig: go.Figure, node_info: Dict, smry: pd.DataFrame, real: int, chart_row: int
) -> None:
    # pylint: disable=too-many-locals
    """Adding area traces to the single realization control mode bar"""
    sumvec = node_info["ctrlmode_sumvec"]
    categories = get_ctrlmode_categories(node_info["type"])
    df = smry[smry.REAL == real]
    prev_ctrlmode = df[df.DATE == df.DATE.min()][sumvec].values[0]
    ctrlmode_startdate = df.DATE.min()
    end_date = df.DATE.max()
    start_date = df.DATE.min()

    for _, row in df.iterrows():
        ctrlmode = row[sumvec]
        if ctrlmode != prev_ctrlmode or row.DATE == end_date:
            category = categories[str(prev_ctrlmode)]
            add_area_trace(
                fig,
                [
                    start_date,
                    ctrlmode_startdate,
                    ctrlmode_startdate,
                    row.DATE,
                    row.DATE,
                    end_date,
                ],
                [0, 0, 1, 1, 0, 0],
                category["name"],
                category["color"],
                row=chart_row,
                showlegend=False,
                linewidth=0,
            )
            ctrlmode_startdate = row.DATE
        prev_ctrlmode = ctrlmode


def get_filtered_smry(
    node_network: dict,
    ctrlmode_sumvec: str,
    pressure_plot_mode: PressurePlotMode,
    real: int,
    smry: pd.DataFrame,
) -> pd.DataFrame:
    """Filters the summary dataframe according to the valid
    start and end dates for the network. Removing un-necessary rows/columns
    and averaging over realizations if that option is selected.
    """
    start_date = pd.to_datetime(node_network["start_date"])
    end_date = pd.to_datetime(node_network["end_date"])
    sumvecs = ["REAL", "DATE", ctrlmode_sumvec] + [
        nodedict["pressure"]
        for nodedict in node_network["nodes"]
        if nodedict["pressure"] in smry.columns
    ]
    df = smry[smry.DATE >= start_date][sumvecs]
    if end_date is not None:
        df = df[df.DATE < end_date]

    if pressure_plot_mode == PressurePlotMode.MEAN:
        # Filter out realizations whitout production
        df = df[df[ctrlmode_sumvec] != 0.0]
        # Group by date and take the mean of each group
        df = df.groupby("DATE").mean().reset_index()
    elif pressure_plot_mode == PressurePlotMode.SINGLE_REAL:
        df = df[df.REAL == real]
    return df


def add_area_trace(
    fig: go.Figure,
    x_series: list,
    y_series: list,
    name: str,
    color: str,
    row: int,
    showlegend: bool = True,
    linewidth: float = 0.2,
) -> go.Figure:
    """Adding single area trace to subplot"""
    fig.add_trace(
        go.Scatter(
            x=x_series,
            y=y_series,
            hoverinfo="text+x+y",
            hoveron="fills",
            mode="lines",
            # fill="tonexty",
            # fillcolor=color,
            line=dict(width=linewidth, color=color),
            name=name,
            text=name,
            stackgroup="one",
            legendgroup="Ctrl Modes",
            showlegend=showlegend,
        ),
        row=row,
        col=1,
    )


def get_ctrlmode_categories(node_type: NodeType) -> dict:
    """Returning name and color for the control mode values"""
    if node_type == NodeType.WELL:
        return {
            "0.0": {"name": "SHUT/STOP", "color": "#302f2f"},  # grey
            "1.0": {"name": "ORAT", "color": StandardColors.OIL_GREEN.value},  # green
            "2.0": {"name": "WRAT", "color": StandardColors.WATER_BLUE.value},  # blue
            "3.0": {"name": "GRAT", "color": StandardColors.GAS_RED.value},  # red
            "4.0": {"name": "LRAT", "color": "#b06d15"},  # muted purple
            "5.0": {"name": "RESV", "color": "#67ab99"},  # green/blue
            "6.0": {"name": "THP", "color": "#7e5980"},  # purple
            "7.0": {"name": "BHP", "color": "#1f77b4"},  # muted blue
            "-1.0": {"name": "GRUP", "color": "#cfcc74"},  # yellow
            "Interpolated": {"name": "Interpolated", "color": "#ffffff"},  # white
            "Other": {"name": "Other", "color": "#ffffff"},  # white
        }
    if node_type == NodeType.GROUP:
        return {
            "0.0": {"name": "NONE", "color": "#302f2f"},  # grey
            "1.0": {"name": "ORAT", "color": StandardColors.OIL_GREEN.value},  # green
            "2.0": {"name": "WRAT", "color": StandardColors.WATER_BLUE.value},  # blue
            "3.0": {"name": "GRAT", "color": StandardColors.GAS_RED.value},  # red
            "4.0": {"name": "LRAT", "color": "#b06d15"},  # muted purple
            "5.0": {"name": "RESV", "color": "#67ab99"},  # green/blue
            "6.0": {"name": "PRBL", "color": "#7e5980"},  # purple
            "7.0": {"name": "ENERGY", "color": "#1f77b4"},  # muted blue
            "-ve": {"name": "GRUP", "color": "#cfcc74"},  # yellow
            "Interpolated": {"name": "Interpolated", "color": "#ffffff"},  # white
            "Other": {"name": "Other", "color": "#ffffff"},  # white
        }
    raise ValueError(f"Node type: {node_type} not implemented")



from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import pandas as pd
import webviz_core_components as wcc
from dash import Input, Output, dcc, html
from plotly.subplots import make_subplots
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.utils import calculate_slider_step
from webviz_config.webviz_store import webvizstore

import webviz_subsurface._utils.parameter_response as parresp
from webviz_subsurface._datainput.fmu_input import load_csv
from webviz_subsurface._figures import create_figure
from webviz_subsurface._models import ParametersModel
from webviz_subsurface._providers import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
    EnsembleTableProvider,
    EnsembleTableProviderFactory,
    Frequency,
    get_matching_vector_names,
)


class ParameterResponseCorrelation(WebvizPluginABC):
    """Visualizes correlations between numerical input parameters and responses.

---
**Three main options for input data: Aggregated, file per realization and read from `.arrow`.**

**Using aggregated data**
* **`parameter_csv`:** Aggregated csvfile for input parameters with `REAL` and `ENSEMBLE` columns \
(absolute path or relative to config file).
* **`response_csv`:** Aggregated csvfile for response parameters with `REAL` and `ENSEMBLE` \
columns (absolute path or relative to config file).


**Using a response file per realization**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`response_file`:** Local (per realization) csv file for response parameters (Cannot be \
                    combined with `response_csv` and `parameter_csv`).


**Using simulation time series data directly from `UNSMRY` files as responses**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize. The lack of `response_file` \
                implies that the input data should be time series data from `.arrow` file \
* **`rel_file_pattern`:** Relative file path to `.arrow` file.
* **`column_keys`:** (Optional) slist of simulation vectors to include as responses when reading \
                from UNSMRY-files in the defined ensembles (default is all vectors). * can be \
                used as wild card.
* **`sampling`:** (Optional) sampling frequency when reading simulation data directly from \
               `.UNSMRY`-files (default is monthly).

?> The `.arrow` input method implies that the "DATE" vector will be used as a filter \
   of type `single` (as defined below under `response_filters`).


**Common settings for all input options**

All of these are optional, some have defaults seen in the code snippet below.

* **`response_filters`:** Optional dictionary of responses (columns in csv file or simulation \
                       vectors) that can be used as row filtering before aggregation. \
                       Valid options:
    * `single`: Dropdown with single selection.
    * `multi`: Dropdown with multiple selection.
    * `range`: Slider with range selection.
* **`response_ignore`:** List of response (columns in csv or simulation vectors) to ignore \
                      (cannot use with response_include).
* **`response_include`:** List of response (columns in csv or simulation vectors) to include \
                       (cannot use with response_ignore).
* **`aggregation`:** Initial way to aggregate responses per realization. Either `sum` or `mean`.
* **`corr_method`:** Initial correlation method. Either `pearson` or `spearman`.

---

?> Non-numerical (string-based) input parameters and responses are removed.

?> The responses will be aggregated per realization; meaning that if your filters do not reduce \
the response to a single value per realization in your data, the values will be aggregated \
accoording to your defined `aggregation`. If e.g. the response is a form of volume, \
and the filters are regions (or other subdivisions of the total volume), then `sum` would \
be a natural aggregation. If on the other hand the response is the pressures in the \
same volume, aggregation as `mean` over the subdivisions of the same volume \
would make more sense (though the pressures in this case would not be volume weighted means, \
and the aggregation would therefore likely be imprecise).

!> It is **strongly recommended** to keep the data frequency to a regular frequency (like \
`monthly` or `yearly`). This applies to both csv input and when reading from `.arrow` \
(controlled by the `sampling` key). This is because the statistics are calculated per DATE over \
all realizations in an ensemble, and the available dates should therefore not differ between \
individual realizations of an ensemble.

**Using aggregated data**

The `parameter_csv` file must have columns `REAL`, `ENSEMBLE` and the parameter columns.

The `response_csv` file must have columns `REAL`, `ENSEMBLE` and the response columns \
(and the columns to use as `response_filters`, if that option is used).


**Using a response file per realization**

Parameters are extracted automatically from the `parameters.txt` files in the individual
realizations, using the `fmu-ensemble` library.

The `response_file` must have the response columns (and the columns to use as `response_filters`, \
if that option is used).


**Using simulation time series data directly from `.arrow` files as responses**

Parameters are extracted automatically from the `parameters.txt` files in the individual
realizations.

Responses are extracted automatically from the `.arrow` files in the individual realizations.
"""

    # pylint:disable=too-many-arguments, too-many-locals
    def __init__(
        self,
        app,
        webviz_settings: WebvizSettings,
        parameter_csv: Path = None,
        response_csv: Path = None,
        ensembles: list = None,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        response_file: str = None,
        response_filters: dict = None,
        response_ignore: list = None,
        response_include: list = None,
        column_keys: list = None,
        sampling: str = "monthly",
        aggregation: str = "sum",
        corr_method: str = "pearson",
    ):

        super().__init__()

        self.parameter_csv = parameter_csv if parameter_csv else None
        self.response_csv = response_csv if response_csv else None
        self.response_file = response_file if response_file else None
        self.response_filters = response_filters if response_filters else {}
        self.column_keys = column_keys
        self._sampling = Frequency(sampling)
        self.corr_method = corr_method
        self.aggregation = aggregation
        if response_ignore and response_include:
            raise ValueError(
                'Incorrent argument. either provide "response_include", '
                '"response_ignore" or neither'
            )
        if parameter_csv and response_csv:
            if ensembles or response_file:
                raise ValueError(
                    'Incorrect arguments. Either provide "csv files" or '
                    '"ensembles and response_file".'
                )
            parameterdf = read_csv(self.parameter_csv)
            self.responsedf = read_csv(self.response_csv)

        elif ensembles:
            self.ens_paths = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }
            table_provider_factory = EnsembleTableProviderFactory.instance()

            parameterdf = create_df_from_table_provider(
                provider_set={
                    ens_name: table_provider_factory.create_from_per_realization_parameter_file(
                        ens_path
                    )
                    for ens_name, ens_path in self.ens_paths.items()
                }
            )

            if self.response_file:
                self.responsedf = load_csv(
                    ensemble_paths=self.ens_paths,
                    csv_file=response_file,
                    ensemble_set_name="EnsembleSet",
                )
            else:
                smry_provider_factory = EnsembleSummaryProviderFactory.instance()
                provider_set = {
                    ens_name: smry_provider_factory.create_from_arrow_unsmry_presampled(
                        ens_path, rel_file_pattern, self._sampling
                    )
                    for ens_name, ens_path in self.ens_paths.items()
                }
                self.response_filters["DATE"] = "single"
                self.responsedf = create_df_from_summary_provider(
                    provider_set,
                    self.column_keys,
                )
        else:
            raise ValueError(
                'Incorrect arguments. Either provide "csv files" or "ensembles and response_file".'
            )
        pmodel = ParametersModel(
            dataframe=parameterdf, keep_numeric_only=True, drop_constants=True
        )
        self.parameterdf = pmodel.dataframe
        self.parameter_columns = pmodel.parameters

        parresp.check_runs(self.parameterdf, self.responsedf)
        parresp.check_response_filters(self.responsedf, self.response_filters)

        # Only select numerical responses
        self.response_columns = parresp.filter_numerical_columns(
            df=self.responsedf,
            column_ignore=response_ignore,
            column_include=response_include,
            filter_columns=self.response_filters.keys(),
        )

        self.theme = webviz_settings.theme
        self.set_callbacks(app)

    @property
    def tour_steps(self) -> List[Dict[str, Any]]:
        steps = [
            {
                "id": self.uuid("layout"),
                "content": (
                    "Dashboard displaying correlation between selected "
                    "response and input parameters."
                ),
            },
            {
                "id": self.uuid("correlation-graph"),
                "content": (
                    "Visualization of the correlations between currently selected "
                    "response and input parameters ranked by the absolute correlation "
                    "coefficient. Click on any correlation to visualize the distribution "
                    "between that parameter and the response."
                ),
            },
            {
                "id": self.uuid("distribution-graph"),
                "content": (
                    "Visualization of the distribution of the response and the selected "
                    "input parameter in the correlation chart."
                ),
            },
            {
                "id": self.uuid("ensemble"),
                "content": ("Select the active ensemble."),
            },
            {
                "id": self.uuid("responses"),
                "content": ("Select the active response."),
            },
            {
                "id": self.uuid("correlation-method"),
                "content": ("Select Pearson or Spearman correlation."),
            },
            {
                "id": self.uuid("aggregation"),
                "content": (
                    "Select whether the response after filtering should be aggregated "
                    "by summation or mean."
                ),
            },
            {
                "id": self.uuid("correlation-cutoff"),
                "content": (
                    "Slider to set a minimum correlation factor for parameters shown "
                    "in plots."
                ),
            },
            {
                "id": self.uuid("max-params"),
                "content": ("Slider to set a maximum number of parameters shown"),
            },
            {
                "id": self.uuid("filters"),
                "content": ("Filters for response and parameters to correlate with."),
            },
        ]

        return steps

    @property
    def ensembles(self) -> List[str]:
        """Returns list of ensembles"""
        return list(self.parameterdf["ENSEMBLE"].unique())

    @property
    def filter_layout(self) -> List[Any]:
        """Layout to display selectors for response filters"""
        children = []
        for col_name, col_type in self.response_filters.items():
            domid = self.uuid(f"filter-{col_name}")
            values = list(self.responsedf[col_name].unique())
            if col_type == "multi":
                selector = wcc.SelectWithLabel(
                    label=f"{col_name}:",
                    id=domid,
                    options=[{"label": val, "value": val} for val in values],
                    value=values,
                    multi=True,
                    size=min(10, len(values)),
                    collapsible=True,
                )
            elif col_type == "single":
                selector = wcc.Dropdown(
                    label=f"{col_name}:",
                    id=domid,
                    options=[{"label": val, "value": val} for val in values],
                    value=values[-1],
                    multi=False,
                    clearable=False,
                )
            elif col_type == "range":
                selector = make_range_slider(domid, self.responsedf[col_name], col_name)
            else:
                return children
            children.append(selector)

        children.append(
            wcc.SelectWithLabel(
                label="Parameters:",
                id=self.uuid("parameter-filter"),
                options=[
                    {"label": val, "value": val} for val in self.parameter_columns
                ],
                value=self.parameter_columns,
                multi=True,
                size=min(10, len(self.parameter_columns)),
                collapsible=True,
            )
        )

        return children

    @property
    def control_layout(self) -> List[Any]:
        """Layout to select e.g. iteration and response"""
        max_params = len(self.parameter_columns)
        return [
            wcc.Dropdown(
                label="Ensemble:",
                id=self.uuid("ensemble"),
                options=[{"label": ens, "value": ens} for ens in self.ensembles],
                clearable=False,
                value=self.ensembles[0],
            ),
            wcc.Dropdown(
                label="Response:",
                id=self.uuid("responses"),
                options=[{"label": col, "value": col} for col in self.response_columns],
                clearable=False,
                value=self.response_columns[0],
            ),
            wcc.RadioItems(
                label="Correlation method:",
                id=self.uuid("correlation-method"),
                options=[
                    {"label": opt.capitalize(), "value": opt}
                    for opt in ["pearson", "spearman"]
                ],
                vertical=False,
                value=self.corr_method,
            ),
            wcc.RadioItems(
                label="Response aggregation:",
                id=self.uuid("aggregation"),
                options=[
                    {"label": opt.capitalize(), "value": opt} for opt in ["sum", "mean"]
                ],
                vertical=False,
                value=self.aggregation,
            ),
            html.Div(
                wcc.Slider(
                    label="Correlation cut-off (abs):",
                    id=self.uuid("correlation-cutoff"),
                    min=0,
                    max=1,
                    step=0.1,
                    marks={"0": 0, "1": 1},
                    value=0,
                ),
                style={"margin-top": "10px"},
            ),
            html.Div(
                wcc.Slider(
                    label="Max number of parameters:",
                    id=self.uuid("max-params"),
                    min=1,
                    max=max_params,
                    step=1,
                    marks={1: "1", max_params: str(max_params)},
                    value=max_params,
                ),
                style={"marginTop": "10px"},
            ),
        ]

    @property
    def layout(self) -> wcc.FlexBox:
        """Main layout"""
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                wcc.FlexColumn(
                    flex=1,
                    children=wcc.Frame(
                        style={"height": "80vh"},
                        children=[
                            wcc.Selectors(
                                label="Controls", children=self.control_layout
                            ),
                            wcc.Selectors(
                                label="Filters",
                                id=self.uuid("filters"),
                                children=self.filter_layout,
                            ),
                        ],
                    ),
                ),
                wcc.FlexColumn(
                    flex=3,
                    children=wcc.Frame(
                        color="white",
                        highlight=False,
                        style={"height": "80vh"},
                        children=[
                            wcc.Graph(
                                style={"height": "75vh"},
                                id=self.uuid("correlation-graph"),
                            ),
                            dcc.Store(
                                id=self.uuid("initial-parameter"),
                                storage_type="session",
                            ),
                        ],
                    ),
                ),
                wcc.FlexColumn(
                    flex=3,
                    children=wcc.Frame(
                        color="white",
                        highlight=False,
                        style={"height": "80vh"},
                        children=wcc.Graph(
                            style={"height": "75vh"}, id=self.uuid("distribution-graph")
                        ),
                    ),
                ),
            ],
        )

    @property
    def correlation_input_callbacks(self) -> List[Input]:
        """List of Inputs for correlation callback"""
        callbacks = [
            Input(self.uuid("ensemble"), "value"),
            Input(self.uuid("responses"), "value"),
            Input(self.uuid("max-params"), "value"),
            Input(self.uuid("parameter-filter"), "value"),
            Input(self.uuid("correlation-method"), "value"),
            Input(self.uuid("aggregation"), "value"),
            Input(self.uuid("correlation-cutoff"), "value"),
        ]
        if self.response_filters:
            for col_name in self.response_filters:
                callbacks.append(Input(self.uuid(f"filter-{col_name}"), "value"))
        return callbacks

    @property
    def distribution_input_callbacks(self) -> List[Input]:
        """List of Inputs for distribution callback"""
        callbacks = [
            Input(self.uuid("correlation-graph"), "clickData"),
            Input(self.uuid("initial-parameter"), "data"),
            Input(self.uuid("ensemble"), "value"),
            Input(self.uuid("responses"), "value"),
            Input(self.uuid("aggregation"), "value"),
        ]
        if self.response_filters:
            for col_name in self.response_filters:
                callbacks.append(Input(self.uuid(f"filter-{col_name}"), "value"))
        return callbacks

    def set_callbacks(self, app) -> None:
        @app.callback(
            [
                Output(self.uuid("correlation-graph"), "figure"),
                Output(self.uuid("initial-parameter"), "data"),
            ],
            self.correlation_input_callbacks,
        )
        def _update_correlation_graph(
            ensemble: str,
            response: str,
            max_parameters: int,
            selected_parameters: List[str],
            correlation_method: str,
            aggregation: str,
            correlation_cutoff: float,
            *filters,
        ):
            """Callback to update correlation graph

            1. Filters and aggregates response dataframe per realization
            2. Filters parameters dataframe on selected ensemble
            3. Merge parameter and response dataframe
            4. Correlate merged dataframe
            5. Sort correlation for selected response by absolute values
            6. Remove nan values return correlation graph
            """

            filteroptions = parresp.make_response_filters(
                response_filters=self.response_filters,
                response_filter_values=filters,
            )
            responsedf = parresp.filter_and_sum_responses(
                self.responsedf,
                ensemble,
                response,
                filteroptions=filteroptions,
                aggregation=aggregation,
            )
            parameterdf = self.parameterdf[
                ["ENSEMBLE", "REAL"] + selected_parameters
            ].loc[self.parameterdf["ENSEMBLE"] == ensemble]

            df = pd.merge(responsedf, parameterdf, on=["REAL"])
            corrdf = correlate(df, response=response, method=correlation_method)
            try:
                corr_response = (
                    corrdf[response]
                    .dropna()
                    .drop(["REAL", response], axis=0)
                    .tail(n=max_parameters)
                )
                corr_response = corr_response[corr_response.abs() >= correlation_cutoff]
                return (
                    make_correlation_plot(
                        corr_response,
                        response,
                        self.theme,
                        correlation_method,
                        correlation_cutoff,
                        max_parameters,
                    ),
                    corr_response.index[-1],
                )
            except (KeyError, ValueError):
                return (
                    {
                        "layout": {
                            "title": "<b>Cannot calculate correlation for given selection</b><br>"
                            "Select a different response or filter setting."
                        }
                    },
                    None,
                )

        @app.callback(
            Output(self.uuid("distribution-graph"), "figure"),
            self.distribution_input_callbacks,
        )
        def _update_distribution_graph(
            clickdata, initial_parameter, ensemble, response, aggregation, *filters
        ):
            """Callback to update distribution graphs.

            1. Filters and aggregates response dataframe per realization
            2. Filters parameters dataframe on selected ensemble
            3. Merge parameter and response dataframe
            4. Generate scatterplot and histograms
            """
            if clickdata:
                parameter = clickdata["points"][0]["y"]
            elif initial_parameter:
                parameter = initial_parameter
            else:
                return {}
            filteroptions = parresp.make_response_filters(
                response_filters=self.response_filters,
                response_filter_values=filters,
            )
            responsedf = parresp.filter_and_sum_responses(
                self.responsedf,
                ensemble,
                response,
                filteroptions=filteroptions,
                aggregation=aggregation,
            )
            parameterdf = self.parameterdf.loc[self.parameterdf["ENSEMBLE"] == ensemble]
            df = pd.merge(responsedf, parameterdf, on=["REAL"])[
                ["REAL", parameter, response]
            ]
            return make_distribution_plot(df, parameter, response, self.theme)

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        if self.parameter_csv and self.response_csv:
            return [
                (
                    read_csv,
                    [
                        {
                            "csv_file": self.parameter_csv,
                        }
                    ],
                ),
                (
                    read_csv,
                    [
                        {
                            "csv_file": self.response_csv,
                        }
                    ],
                ),
            ]
        if self.response_file:
            return [
                (
                    load_csv,
                    [
                        {
                            "ensemble_paths": self.ens_paths,
                            "csv_file": self.response_file,
                            "ensemble_set_name": "EnsembleSet",
                        }
                    ],
                )
            ]
        return []


def correlate(inputdf, response, method="pearson") -> pd.DataFrame:
    """Returns the correlation matrix for a dataframe"""
    if method == "pearson":
        corrdf = inputdf.corr(method=method)
    elif method == "spearman":
        corrdf = inputdf.rank().corr(method="pearson")
    else:
        raise ValueError(
            f"Correlation method {method} is invalid. "
            "Available methods are 'pearson' and 'spearman'"
        )
    return corrdf.reindex(corrdf[response].abs().sort_values().index)


def make_correlation_plot(
    series, response, theme, corr_method, corr_cutoff, max_parameters
) -> Dict[str, Any]:
    """Make Plotly trace for correlation plot"""
    xaxis_range = max(abs(series.values)) * 1.1
    layout = {
        "barmode": "relative",
        "margin": {"l": 200, "r": 50, "b": 20, "t": 100},
        "xaxis": {"range": [-xaxis_range, xaxis_range]},
        "yaxis": {"dtick": 1},
        "title": (
            f"Correlations between {response} and input parameters<br>"
            f"{corr_method.capitalize()} correlation with abs cut-off {corr_cutoff}"
            f" and max {max_parameters} parameters"
        ),
    }
    layout = theme.create_themed_layout(layout)

    return {
        "data": [
            {"x": series.values, "y": series.index, "orientation": "h", "type": "bar"}
        ],
        "layout": layout,
    }


def make_distribution_plot(df, parameter, response, theme):
    """Make plotly traces for scatterplot and histograms for selected
    response and input parameter"""

    fig = make_subplots(
        rows=4,
        cols=2,
        specs=[
            [{"colspan": 2, "rowspan": 2}, None],
            [None, None],
            [{"rowspan": 2}, {"rowspan": 2}],
            [None, None],
        ],
    )
    scatter_trace, trendline = create_figure(
        plot_type="scatter",
        data_frame=df,
        x=parameter,
        y=response,
        trendline="ols",
        hover_data={"REAL": True},
        color_discrete_sequence=["SteelBlue"],
        marker={"size": 20, "opacity": 0.7},
    ).data

    fig.add_trace(scatter_trace, 1, 1)
    fig.add_trace(trendline, 1, 1)
    fig.add_trace(
        {
            "type": "histogram",
            "x": df[parameter],
            "showlegend": False,
        },
        3,
        1,
    )
    fig.add_trace(
        {
            "type": "histogram",
            "x": df[response],
            "showlegend": False,
        },
        3,
        2,
    )
    fig["layout"].update(
        theme_layout(
            theme.plotly_theme,
            {
                "bargap": 0.05,
                "xaxis": {
                    "title": parameter,
                },
                "yaxis": {"title": response},
                "xaxis2": {"title": parameter},
                "xaxis3": {"title": response},
                "title": f"Distribution of {response} and {parameter}",
            },
        )
    )
    return fig


def make_range_slider(domid, values, col_name) -> wcc.RangeSlider:
    try:
        values.apply(pd.to_numeric, errors="raise")
    except ValueError as exc:
        raise ValueError(
            f"Cannot calculate filter range for {col_name}. "
            "Ensure that it is a numerical column."
        ) from exc
    return wcc.RangeSlider(
        label=f"{col_name}:",
        id=domid,
        min=values.min(),
        max=values.max(),
        step=calculate_slider_step(
            min_value=values.min(),
            max_value=values.max(),
            steps=len(list(values.unique())) - 1,
        ),
        value=[values.min(), values.max()],
        marks={
            str(values.min()): {"label": f"{values.min():.2f}"},
            str(values.max()): {"label": f"{values.max():.2f}"},
        },
    )


def theme_layout(theme, specific_layout) -> Dict:
    layout = {}
    layout.update(theme["layout"])
    layout.update(specific_layout)
    return layout


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=False)


def create_df_from_table_provider(
    provider_set: Dict[str, EnsembleTableProvider]
) -> pd.DataFrame:
    """Aggregates parameters from all ensemble into a common dataframe."""
    dfs = []
    for ens_name, provider in provider_set.items():
        df = provider.get_column_data(column_names=provider.column_names())
        df["ENSEMBLE"] = ens_name
        dfs.append(df)
    return pd.concat(dfs)


def create_df_from_summary_provider(
    provider_set: Dict[str, EnsembleSummaryProvider], column_keys: List[str]
) -> pd.DataFrame:
    """Aggregates summary data from all ensembles into a common dataframe."""
    dfs = []
    for ens_name, provider in provider_set.items():
        if column_keys is None:
            matching_sumvecs = provider.vector_names()
        else:
            matching_sumvecs = get_matching_vector_names(provider, column_keys)
        if not matching_sumvecs:
            raise ValueError(
                f"No vectors matching the given column_keys: {column_keys} for ensemble: {ens_name}"
            )
        df = provider.get_vectors_df(matching_sumvecs, None)
        df["ENSEMBLE"] = ens_name
        dfs.append(df)

    df_all = pd.concat(dfs)
    df_all["DATE"] = pd.to_datetime(df_all["DATE"]).dt.strftime("%Y-%m-%d")
    return df_all



from typing import Callable, List, Tuple, Type

from dash.development.base_component import Component
from webviz_config import WebvizPluginABC, WebvizSettings

from ._error import error
from ._plugin_ids import PluginIds
from .views import ParameterPlot
from .views.parameter_plot._parameter_plot import get_parameters


class ParameterCorrelation(WebvizPluginABC):
    """Showing parameter correlations using a correlation matrix,
    and scatter plot for any given pair of parameters.

    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`drop_constants`:** Drop constant parameters.

    ---
    Parameter values are extracted automatically from the `parameters.txt` files in the individual
    realizations of your defined `ensembles`, using the `fmu-ensemble` library."""

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: list,
        drop_constants: bool = True,
    ) -> None:
        super().__init__(stretch=True)

        self.error_message = ""

        try:
            self.ensembles = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.plotly_theme = webviz_settings.theme.plotly_theme
        except TypeError:
            self.error_message = "WebvizSettings not iterable"
            self.ensembles = {"iter-0": "iter-0", "iter-3": "iter-3"}
        except AttributeError:
            self.error_message = "'Dash' object has no attribute 'theme'"

        self.plot = ParameterPlot(self.ensembles, webviz_settings, drop_constants)

        self.add_view(
            self.plot,
            PluginIds.ParaCorrGroups.PARACORR,
            PluginIds.ParaCorrGroups.GROUPNAME,
        )

    @property
    def layout(self) -> Type[Component]:
        return error(self.error_message)

    @property
    def tour_steps(self) -> List[dict]:
        """Tour of the plugin"""
        return self.plot.tour_steps

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (get_parameters, [{"ensemble_path": v} for v in self.ensembles.values()])
        ]



from typing import List

import numpy as np
import pandas as pd

from webviz_subsurface._utils.dataframe_utils import (
    assert_date_column_is_datetime_object,
    make_date_column_datetime_object,
)

from .._types import StatisticsOptions


def create_vectors_statistics_df(vectors_df: pd.DataFrame) -> pd.DataFrame:
    """
    Create vectors statistics dataframe for given vectors in columns of provided vectors dataframe

    Calculate min, max, mean, p10, p90 and p50 for each vector in dataframe column

    `Input:`
    * vectors_df: pd.DataFrame - Dataframe with vectors dataframe and columns:
        ["DATE", "REAL", vector1, ... , vectorN]

    `Returns:`
    * Dataframe with double column level:\n
      [ "DATE",     vector1,                        ... vectorN
                    MEAN, MIN, MAX, P10, P90, P50   ... MEAN, MIN, MAX, P10, P90, P50]
    """
    assert_date_column_is_datetime_object(vectors_df)

    # Get vectors names, keep order
    columns_list = list(vectors_df.columns)
    vector_names = sorted(
        (set(columns_list) ^ set(["DATE", "REAL"])), key=columns_list.index
    )

    # If no rows of data:
    if not vectors_df.shape[0]:
        columns_tuples = [("DATE", "")]
        for vector in vector_names:
            columns_tuples.extend(
                [
                    (vector, StatisticsOptions.MEAN),
                    (vector, StatisticsOptions.MIN),
                    (vector, StatisticsOptions.MAX),
                    (vector, StatisticsOptions.P10),
                    (vector, StatisticsOptions.P90),
                    (vector, StatisticsOptions.P50),
                ]
            )
        return pd.DataFrame(columns=pd.MultiIndex.from_tuples(columns_tuples))

    # Invert p10 and p90 due to oil industry convention.
    def p10(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=90)

    def p90(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=10)

    def p50(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=50)

    statistics_df: pd.DataFrame = (
        vectors_df[["DATE"] + vector_names]
        .groupby(["DATE"])
        .agg([np.nanmean, np.nanmin, np.nanmax, p10, p90, p50])
        .reset_index(level=["DATE"], col_level=0)
    )

    # Rename columns to StatisticsOptions enum types for strongly typed format
    col_stat_label_map = {
        "nanmin": StatisticsOptions.MIN,
        "nanmax": StatisticsOptions.MAX,
        "nanmean": StatisticsOptions.MEAN,
        "p10": StatisticsOptions.P10,
        "p90": StatisticsOptions.P90,
        "p50": StatisticsOptions.P50,
    }
    statistics_df.rename(columns=col_stat_label_map, level=1, inplace=True)

    make_date_column_datetime_object(statistics_df)

    return statistics_df



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import ChartType


class WellOverviewChartType(SettingsGroupABC):
    class Ids(StrEnum):
        CHARTTYPE = "charttype"

    def __init__(self) -> None:
        super().__init__("Chart Type")

    def layout(self) -> List[Component]:
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(self.Ids.CHARTTYPE),
                options=[
                    {"label": "Bar chart", "value": ChartType.BAR},
                    {"label": "Pie chart", "value": ChartType.PIE},
                    {"label": "Stacked area chart", "value": ChartType.AREA},
                ],
                value=ChartType.BAR,
                vertical=True,
            )
        ]



import hashlib
import logging
import os
from pathlib import Path

from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WebvizRunMode

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._fault_polygons_discovery import discover_per_realization_fault_polygons_files
from ._provider_impl_file import ProviderImplFile
from .ensemble_fault_polygons_provider import EnsembleFaultPolygonsProvider

LOGGER = logging.getLogger(__name__)


class EnsembleFaultPolygonsProviderFactory(WebvizFactory):
    def __init__(self, root_storage_folder: Path, allow_storage_writes: bool) -> None:
        self._storage_dir = Path(root_storage_folder) / __name__
        self._allow_storage_writes = allow_storage_writes

        LOGGER.info(
            f"EnsembleFaultPolygonsProviderFactory init: storage_dir={self._storage_dir}"
        )

        if self._allow_storage_writes:
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "EnsembleFaultPolygonsProviderFactory":
        """Static method to access the singleton instance of the factory."""

        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(
            EnsembleFaultPolygonsProviderFactory
        )
        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE

            factory = EnsembleFaultPolygonsProviderFactory(storage_folder, allow_writes)

            # Store the factory object in the global factory registry
            WEBVIZ_FACTORY_REGISTRY.set_factory(
                EnsembleFaultPolygonsProviderFactory, factory
            )

        return factory

    def create_from_ensemble_fault_polygons_files(
        self, ens_path: str
    ) -> EnsembleFaultPolygonsProvider:
        timer = PerfTimer()

        storage_key = f"ens__{_make_hash_string(ens_path)}"
        provider = ProviderImplFile.from_backing_store(self._storage_dir, storage_key)
        if provider:
            LOGGER.info(
                f"Loaded fault polygons provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load fault polygons provider for {ens_path}")

        LOGGER.info(f"Importing/copying fault polygons data for: {ens_path}")

        timer.lap_s()
        sim_fault_polygons_files = discover_per_realization_fault_polygons_files(
            ens_path
        )

        et_discover_s = timer.lap_s()

        ProviderImplFile.write_backing_store(
            self._storage_dir,
            storage_key,
            sim_fault_polygons=sim_fault_polygons_files,
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplFile.from_backing_store(self._storage_dir, storage_key)
        if not provider:
            raise ValueError(
                f"Failed to load/create fault polygons provider for {ens_path}"
            )

        LOGGER.info(
            f"Saved fault polygons provider to backing store in {timer.elapsed_s():.2f}s ("
            f"discover={et_discover_s:.2f}s, write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec



import datetime
import glob
import logging
import os
import re
from pathlib import Path
from typing import Optional

import datacompy  # pylint: disable=import-error, useless-suppression
import dateutil.parser  # type: ignore
import ecl2df
import numpy as np
import pandas as pd
from fmu.ensemble import ScratchEnsemble

from .ensemble_summary_provider import Frequency
from .ensemble_summary_provider_factory import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
)

# pylint: disable=line-too-long

# Note that running this script requires the additional install of datacompy.
#   pip install datacompy

# Tolerances inspired by
#   https://numpy.org/doc/stable/reference/generated/numpy.isclose.html
# and
#   https://pandas.pydata.org/docs/reference/api/pandas.testing.assert_frame_equal.html#pandas.testing.assert_frame_equal
ABS_TOLERANCE = 1e-8
REL_TOLERANCE = 1e-5


def _make_date_column_datetime_object(df: pd.DataFrame) -> pd.DataFrame:

    sampled_date_value = df["DATE"].values[0]

    # Infer datatype based on the first element:
    if isinstance(sampled_date_value, pd.Timestamp):
        df["DATE"] = pd.Series(
            [ts.to_pydatetime() for ts in df["DATE"]], dtype="object"
        )

    elif isinstance(sampled_date_value, np.datetime64):
        df["DATE"] = pd.Series(df["DATE"].dt.to_pydatetime(), dtype="object")

    elif isinstance(sampled_date_value, str):
        # Do not use pd.Series.apply() here, Pandas would try to convert it to
        # datetime64[ns] which is limited at year 2262.
        df["DATE"] = pd.Series(
            [dateutil.parser.parse(datestr) for datestr in df["DATE"]], dtype="object"
        )

    elif isinstance(sampled_date_value, datetime.date):
        df["DATE"] = pd.Series(
            [
                datetime.datetime.combine(dateobj, datetime.datetime.min.time())
                for dateobj in df["DATE"]
            ],
            dtype="object",
        )

    # sampled_after_conv = df["DATE"].values[0]
    # print(f"sampled_after_conv={sampled_after_conv}   type={type(sampled_after_conv)}")

    return df


def _load_smry_dataframe_using_fmu(
    ens_path: str, frequency: Optional[Frequency]
) -> pd.DataFrame:

    time_index: str = "raw"
    if frequency:
        time_index = frequency.value

    print(f"## Loading data into DataFrame using FMU  time_index={time_index}...")

    scratch_ensemble = ScratchEnsemble("tempEnsName", paths=ens_path)
    df = scratch_ensemble.load_smry(time_index=time_index)

    df = _make_date_column_datetime_object(df)

    # Convert float columns to float32 and real column to int32
    floatcols = df.select_dtypes("float").columns
    df[floatcols] = df[floatcols].apply(pd.to_numeric, downcast="float")
    df["REAL"] = df["REAL"].astype("int32")

    # Sort on real, then date to align with provider
    df.sort_values(by=["REAL", "DATE"], inplace=True)
    df.reset_index(drop=True, inplace=True)

    return df


def _load_smry_dataframe_using_ecl2df(
    ens_path: str, frequency: Optional[Frequency]
) -> pd.DataFrame:

    time_index: str = "raw"
    if frequency:
        time_index = frequency.value

    print(f"## Loading data into DataFrame using ECL2DF  time_index={time_index}...")

    realidxregexp = re.compile(r"realization-(\d+)")
    globpattern = os.path.join(ens_path, "eclipse/model/*.UNSMRY")
    globbedpaths = sorted(glob.glob(globpattern))

    per_real_df_arr = []

    for smry_file in globbedpaths:
        real = None
        for path_comp in reversed(smry_file.split(os.path.sep)):
            realmatch = re.match(realidxregexp, path_comp)
            if realmatch:
                real = int(realmatch.group(1))
                break

        if real is None:
            raise ValueError(
                f"Unable to determine realization number for file: {smry_file}"
            )

        print(f"R={real}:  {smry_file}")

        eclfiles = ecl2df.EclFiles(smry_file.replace(".UNSMRY", ""))
        real_df = ecl2df.summary.df(eclfiles, time_index=time_index)
        real_df.insert(0, "REAL", real)
        real_df.index.name = "DATE"
        per_real_df_arr.append(real_df)

    df = pd.concat(per_real_df_arr, sort=False).reset_index()

    df = _make_date_column_datetime_object(df)

    # Convert float columns to float32 and real column to int32
    floatcols = df.select_dtypes("float").columns
    df[floatcols] = df[floatcols].apply(pd.to_numeric, downcast="float")
    df["REAL"] = df["REAL"].astype("int32")

    # Sort on real, then date to align with provider
    df.sort_values(by=["REAL", "DATE"], inplace=True)
    df.reset_index(drop=True, inplace=True)

    return df


def _compare_reference_df_to_provider_get_vectors_df(
    reference_df: pd.DataFrame,
    provider: EnsembleSummaryProvider,
    frequency: Optional[Frequency],
) -> None:

    reference_df = reference_df.reset_index(drop=True)
    # print(ref_df)

    print("## Getting data for all vectors from provider...")
    provider_df = provider.get_vectors_df(provider.vector_names(), frequency)
    provider_df.sort_values(by=["REAL", "DATE"], inplace=True)
    provider_df.reset_index(drop=True, inplace=True)
    # print(provider_df)

    print("## Running compare...")
    compare = datacompy.Compare(
        reference_df,
        provider_df,
        df1_name="ref_df",
        df2_name="provider_df",
        join_columns=["REAL", "DATE"],
        abs_tol=ABS_TOLERANCE,
        rel_tol=REL_TOLERANCE,
    )
    if compare.matches():
        print("## Comparison ok")
    else:
        print("## Comparison DID NOT MATCH")
        print(compare.report())

        print("## Columns that are only present in reference_df:")
        print(compare.df1_unq_columns())

        print("## Columns that are only present in provider_df:")
        print(compare.df2_unq_columns())


def _compare_reference_df_to_provider_get_vectors_for_date(
    reference_df: pd.DataFrame, provider: EnsembleSummaryProvider
) -> None:

    all_dates = reference_df["DATE"].unique()
    num_dates = len(all_dates)
    # print(all_dates)

    lookup_date_arr = []
    lookup_date_arr.append(all_dates[0])
    lookup_date_arr.append(all_dates[int(num_dates / 3)])
    lookup_date_arr.append(all_dates[int(num_dates / 2)])
    lookup_date_arr.append(all_dates[int(2 * num_dates / 3)])
    lookup_date_arr.append(all_dates[num_dates - 1])

    for lookup_date in lookup_date_arr:
        # print(f"lookup_date={lookup_date}   type={type(lookup_date)}")
        ref_res_df = reference_df.loc[reference_df["DATE"] == lookup_date].copy()
        ref_res_df.drop(columns="DATE", inplace=True)
        ref_res_df.reset_index(drop=True, inplace=True)
        # print("## Dumping ref_res_df:")
        # print(ref_res_df)

        provider_df = provider.get_vectors_for_date_df(
            lookup_date, provider.vector_names()
        )
        provider_df.reset_index(drop=True, inplace=True)
        # print("## Dumping provider_df:")
        # print(provider_df)

        print(f"## Running compare for date {lookup_date} ...")
        compare = datacompy.Compare(
            ref_res_df,
            provider_df,
            df1_name="ref_res_df",
            df2_name="provider_df",
            join_columns=["REAL"],
            abs_tol=ABS_TOLERANCE,
            rel_tol=REL_TOLERANCE,
        )
        if compare.matches():
            print(f"## Comparison ok for date {lookup_date}")
        else:
            print(f"## Comparison DID NOT MATCH for date {lookup_date}")
            print(compare.report())

            print("## Columns that are only present in ref_res_df:")
            print(compare.df1_unq_columns())

            print("## Columns that are only present in provider_df:")
            print(compare.df2_unq_columns())


def main() -> None:
    print()
    print("## Running comparison SMRY FMU vs lazy provider")
    print("## =================================================")

    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s %(levelname)-3s [%(name)s]: %(message)s",
    )
    logging.getLogger("webviz_subsurface").setLevel(level=logging.INFO)
    logging.getLogger("webviz_subsurface").setLevel(level=logging.DEBUG)

    root_storage_dir = Path("/home/sigurdp/buf/webviz_storage_dir")

    ensemble_path = "../webviz-subsurface-testdata/01_drogon_ahm/realization-*/iter-0"
    # ensemble_path = (
    #     "../webviz-subsurface-testdata/01_drogon_design/realization-*/iter-0"
    # )
    # ensemble_path = (
    #     "/home/sigurdp/webviz_testdata/reek_history_match_large/realization-*/iter-0"
    # )

    # frequency = None
    # frequency = Frequency.DAILY
    frequency = Frequency.WEEKLY

    print()
    print("## root_storage_dir:", root_storage_dir)
    print("## ensemble_path:", ensemble_path)
    print("## frequency:", frequency)
    print()

    print("## Creating summary provider...")
    factory = EnsembleSummaryProviderFactory(
        root_storage_dir, allow_storage_writes=True
    )
    provider = factory.create_from_arrow_unsmry_lazy(
        ens_path=ensemble_path, rel_file_pattern="share/results/unsmry/*.arrow"
    )
    # provider = factory.create_from_arrow_unsmry_presampled(ensemble_path, frequency)

    print("## Loading data into reference DataFrame...")
    # Note that for version 2.13.0 and earlier of ecl, loading via FMU will not give the
    # correct results. This was remedied in https://github.com/equinor/ecl/pull/837
    # reference_df = _load_smry_dataframe_using_ecl2df(ensemble_path, frequency)
    reference_df = _load_smry_dataframe_using_fmu(ensemble_path, frequency)

    print("## Comparing get_vectors()...")
    resampling_frequency = frequency if provider.supports_resampling() else None
    _compare_reference_df_to_provider_get_vectors_df(
        reference_df, provider, resampling_frequency
    )

    print("## Comparing get_vectors_for date()...")
    _compare_reference_df_to_provider_get_vectors_for_date(reference_df, provider)

    print("## done")


# Running:
# python -m webviz_subsurface._providers.ensemble_summary_provider.dev_compare_fmu_to_lazy_provider
if __name__ == "__main__":
    main()



from typing import Any, List

import webviz_core_components as wcc
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import PressureType


class PressureOption(SettingsGroupABC):
    class Ids(StrEnum):
        PRESSURE_OPTION = "pressure-option"

    def __init__(self) -> None:
        super().__init__("Pressure Option")

    def layout(self) -> List[Any]:
        return [
            wcc.RadioItems(
                id=self.register_component_unique_id(
                    PressureOption.Ids.PRESSURE_OPTION
                ),
                options=[
                    {"label": "BHP", "value": PressureType.BHP},
                    {"label": "DP (BHP-THP)", "value": PressureType.DP},
                ],
                value=PressureType.BHP,
                persistence=True,
                persistence_type="session",
            )
        ]



import warnings

# pylint: disable=no-name-in-module
from webviz_config.plugins import ParameterAnalysis


def test_parameter_analysis(dash_duo, app, shared_settings) -> None:

    plugin = ParameterAnalysis(
        app,
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        column_keys=["WWCT:*"],
        time_index="monthly",
        drop_constants=True,
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    logs = []
    for log in dash_duo.get_logs():
        if "dash_renderer" in log.get("message"):
            warnings.warn(log.get("message"))
        else:
            logs.append(log)
    assert not logs



from typing import List, Optional, Sequence

import pandas as pd

from webviz_subsurface._providers import Frequency

from ....mocks.ensemble_summary_provider_dummy import EnsembleSummaryProviderDummy


class EnsembleSummaryProviderMock(EnsembleSummaryProviderDummy):
    """Mock implementation of EnsembleSummaryProvider for testing derived
    ensemble vectors accessor.

    Implements necessary methods for obtaining wanted test data
    """

    def __init__(self, df: pd.DataFrame) -> None:
        super().__init__()
        self._df = df
        self._vectors: List[str] = [
            elm for elm in df.columns if elm not in ["DATE", "REAL"]
        ]
        self._realizations: List[int] = list(df["REAL"]) if "REAL" in df.columns else []

    #####################################
    #
    # Override methods
    #
    #####################################
    def supports_resampling(self) -> bool:
        return False

    def realizations(self) -> List[int]:
        return self._realizations

    def vector_names(self) -> List[str]:
        return self._vectors

    def get_vectors_df(
        self,
        vector_names: Sequence[str],
        __resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:
        for elm in vector_names:
            if elm not in self._vectors:
                raise ValueError(
                    f'Requested vector "{elm}" not among provider vectors!'
                )
        if realizations:
            # Note: Reset index as providers reset index counter when filtering
            # realization query.
            output = (
                self._df[["DATE", "REAL"] + list(vector_names)]
                .loc[self._df["REAL"].isin(realizations)]
                .reset_index()
            )
            output.drop("index", inplace=True, axis=1)
            return output
        return self._df[["DATE", "REAL"] + list(vector_names)]



import json
import pathlib

_DATA_PATH = pathlib.Path(__file__).parent.absolute() / "abbreviation_data"

VOLUME_TERMINOLOGY = json.loads((_DATA_PATH / "volume_terminology.json").read_text())


def volume_description(column: str) -> str:
    """Return description for the column if defined"""
    try:
        label = VOLUME_TERMINOLOGY[column]["label"]
    except KeyError:
        label = column
    return label


def volume_unit(column: str) -> str:
    """Return unit for the column if defined"""
    try:
        unit = VOLUME_TERMINOLOGY[column]["unit"]
    except KeyError:
        unit = ""
    return unit


def volume_simulation_vector_match(column: str) -> list:
    """Return a list of simulation vectors that match the column
    Useful to verify/propose data to compare.
    """
    try:
        vectors = VOLUME_TERMINOLOGY[column]["eclsum"]
    except KeyError:
        vectors = []
    return vectors if isinstance(vectors, list) else [vectors]



from typing import Dict, List, Optional, Set

import pandas as pd
from plotly.subplots import make_subplots
from webviz_config._theme_class import WebvizConfigTheme

from webviz_subsurface._providers import Frequency
from webviz_subsurface._utils.colors import hex_to_rgb, rgb_to_str, scale_rgb_lightness

from .._types import FanchartOptions, StatisticsOptions
from .._utils.create_vector_traces_utils import (
    create_history_vector_trace,
    create_vector_fanchart_traces,
    create_vector_observation_traces,
    create_vector_realization_traces,
    create_vector_statistics_traces,
    render_hovertemplate,
)
from .graph_figure_builder_base import GraphFigureBuilderBase


class EnsembleSubplotBuilder(GraphFigureBuilderBase):
    """
    Figure builder for creating/building serializable Output property data
    for the callback. Where vector traces are added per ensemble, and subplots
    are categorized per ensemble among selected ensembles.

    Contains functions for adding titles, graph data and retrieving the serialized
    data for callback Output property.

    `Input:`
    * selected_vectors: List[str] - list of selected vector names
    * selected_ensembles: List[str] - list of selected ensemble names
    * vector_colors: dict - Dictionary with vector name as key and graph color as value
    * sampling_frequency: Optional[Frequency] - Sampling frequency of data
    * vector_line_shapes: Dict[str,str] - Dictionary of vector names and line shapes
    * theme: Optional[WebvizConfigTheme] = None - Theme for plugin, given to graph figure
    * line_shape_fallback: str = "linear" - Lineshape fallback
    """

    def __init__(
        self,
        selected_vectors: List[str],
        selected_ensembles: List[str],
        vector_colors: dict,
        sampling_frequency: Optional[Frequency],
        vector_line_shapes: Dict[str, str],
        theme: Optional[WebvizConfigTheme] = None,
        line_shape_fallback: str = "linear",
    ) -> None:
        # Init for base class
        super().__init__()

        self._selected_vectors = selected_vectors
        self._selected_ensembles = selected_ensembles
        self._vector_colors = vector_colors
        self._sampling_frequency = sampling_frequency
        self._line_shape_fallback = line_shape_fallback
        self._vector_line_shapes = vector_line_shapes
        self._history_vector_color = "black"

        # Overwrite graph figure widget
        self._figure = make_subplots(
            rows=max(1, len(self._selected_ensembles)),
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.05,
            subplot_titles=[f'Ensemble: "{elm}"' for elm in self._selected_ensembles],
        )
        if theme:
            self._figure.update_layout(
                theme.create_themed_layout(self._figure.to_dict().get("layout", {}))
            )
        self._set_keep_uirevision()

        # Set for storing added vectors
        self._added_vector_traces: Set[str] = set()

        # Status for added history vectors
        self._added_history_trace = False

    #############################################################################
    #
    # Public methods
    #
    #############################################################################

    def create_graph_legends(self) -> None:
        # Add legends for selected vectors - sort according to selected vectors
        # NOTE: sorted() with key=self._selected_vectors.index requires that all of
        # vectors in self._added_vector_traces set exist in self._selected_vectors list!
        added_vector_traces = sorted(
            self._added_vector_traces, key=self._selected_vectors.index
        )
        for index, vector in enumerate(added_vector_traces, start=1):
            vector_legend_trace = {
                "name": vector,
                "x": [None],
                "y": [None],
                "legendgroup": vector,
                "showlegend": True,
                "visible": True,
                "mode": "lines",
                "line": {
                    "color": self._vector_colors.get(vector, "black"),
                    "shape": self._vector_line_shapes.get(
                        vector, self._line_shape_fallback
                    ),
                },
                "legendrank": index,
            }
            self._figure.add_trace(vector_legend_trace, row=1, col=1)

        # Add legend for history trace with legendrank after vectors
        if self._added_history_trace:
            history_legend_trace = {
                "name": "History",
                "x": [None],
                "y": [None],
                "legendgroup": "History",
                "showlegend": True,
                "visible": True,
                "mode": "lines",
                "line": {
                    "color": self._history_vector_color,
                },
                "legendrank": len(self._added_vector_traces) + 1,
            }
            self._figure.add_trace(
                trace=history_legend_trace,
                row=1,
                col=1,
            )

    def add_realizations_traces(
        self,
        vectors_df: pd.DataFrame,
        ensemble: str,
        color_lightness_scale: Optional[float] = None,
    ) -> None:
        # Dictionary with vector name as key and list of ensemble traces as value
        vector_traces_set: Dict[str, List[dict]] = {}

        # Get vectors - order not important
        vectors: Set[str] = set(vectors_df.columns) - set(["DATE", "REAL"])
        self._validate_vectors_are_selected(vectors)

        for vector in vectors:
            self._added_vector_traces.add(vector)

            vector_df = vectors_df[["DATE", "REAL", vector]]

            color = self._vector_colors.get(vector, "black")
            if color_lightness_scale:
                # Range: 50% - 150% lightness
                scale = max(50.0, min(150.0, color_lightness_scale))
                color = rgb_to_str(scale_rgb_lightness(hex_to_rgb(color), scale))

            line_shape = self._vector_line_shapes.get(vector, self._line_shape_fallback)
            vector_traces_set[vector] = create_vector_realization_traces(
                vector_df=vector_df,
                ensemble=ensemble,
                legend_group=vector,
                color=color,
                line_shape=line_shape,
                hovertemplate=render_hovertemplate(vector, self._sampling_frequency),
            )

        # Add traces to figure
        self._add_vector_traces_set_to_figure(vector_traces_set, ensemble)

    def add_statistics_traces(
        self,
        vectors_statistics_df: pd.DataFrame,
        ensemble: str,
        statistics_options: List[StatisticsOptions],
        line_width: Optional[int] = None,
        color_lightness_scale: Optional[float] = None,
    ) -> None:
        # Dictionary with vector name as key and list of ensemble traces as value
        vector_traces_set: Dict[str, List[dict]] = {}

        # Get vectors - order not important
        vectors: Set[str] = set(
            vectors_statistics_df.columns.get_level_values(0)
        ) - set(["DATE"])
        self._validate_vectors_are_selected(vectors)

        for vector in vectors:
            self._added_vector_traces.add(vector)

            # Retrieve DATE and statistics columns for specific vector
            vector_statistics_df = pd.DataFrame(vectors_statistics_df["DATE"]).join(
                vectors_statistics_df[vector]
            )

            color = self._vector_colors.get(vector, "black")
            if color_lightness_scale:
                # Range: 50% - 150% lightness
                scale = max(50.0, min(150.0, color_lightness_scale))
                color = rgb_to_str(scale_rgb_lightness(hex_to_rgb(color), scale))

            line_shape = self._vector_line_shapes.get(vector, self._line_shape_fallback)
            vector_traces_set[vector] = create_vector_statistics_traces(
                vector_statistics_df=vector_statistics_df,
                color=color,
                legend_group=vector,
                line_shape=line_shape,
                line_width=line_width if line_width else 2,
                statistics_options=statistics_options,
                hovertemplate=render_hovertemplate(vector, self._sampling_frequency),
            )

        # Add traces to figure
        self._add_vector_traces_set_to_figure(vector_traces_set, ensemble)

    def add_fanchart_traces(
        self,
        vectors_statistics_df: pd.DataFrame,
        ensemble: str,
        fanchart_options: List[FanchartOptions],
    ) -> None:
        # Dictionary with vector name as key and list of ensemble traces as value
        vector_traces_set: Dict[str, List[dict]] = {}

        # Get vectors - order not important!
        vectors: Set[str] = set(
            vectors_statistics_df.columns.get_level_values(0)
        ) - set(["DATE"])
        self._validate_vectors_are_selected(vectors)

        for vector in vectors:
            self._added_vector_traces.add(vector)

            # Retrieve DATE and statistics columns for specific vector
            vector_statistics_df = pd.DataFrame(vectors_statistics_df["DATE"]).join(
                vectors_statistics_df[vector]
            )

            color = self._vector_colors.get(vector, "#000000")  # Black Hex color
            line_shape = self._vector_line_shapes.get(vector, self._line_shape_fallback)
            vector_traces_set[vector] = create_vector_fanchart_traces(
                vector_statistics_df=vector_statistics_df,
                hex_color=color,
                legend_group=vector,
                line_shape=line_shape,
                fanchart_options=fanchart_options,
                hovertemplate=render_hovertemplate(vector, self._sampling_frequency),
            )

        # Add traces to figure
        self._add_vector_traces_set_to_figure(vector_traces_set, ensemble)

    def add_history_traces(
        self,
        vectors_df: pd.DataFrame,
        ensemble: str,
    ) -> None:
        if "DATE" not in vectors_df.columns and "REAL" not in vectors_df.columns:
            raise ValueError('vectors_df is missing required columns ["DATE","REAL"]')

        if ensemble is None:
            raise ValueError(
                "Must provide ensemble argument of type str for this implementation!"
            )

        samples = vectors_df["DATE"].tolist()
        vector_trace_set: Dict[str, dict] = {}
        vectors: Set[str] = set(vectors_df.columns) - set(["DATE", "REAL"])
        self._validate_vectors_are_selected(vectors)

        for vector in vectors:
            # Set status for added history trace
            self._added_history_trace = True

            line_shape = self._vector_line_shapes.get(vector, self._line_shape_fallback)
            vector_trace_set[vector] = create_history_vector_trace(
                samples,
                vectors_df[vector].values,
                line_shape=line_shape,
                color=self._history_vector_color,
                vector_name=vector,
            )
        self._add_vector_trace_set_to_figure(vector_trace_set, ensemble)

    def add_vector_observations(
        self, vector_name: str, vector_observations: dict
    ) -> None:
        if vector_name not in self._selected_vectors:
            raise ValueError(f"Vector {vector_name} not among selected vectors!")

        vector_observations_traces_set = {
            vector_name: create_vector_observation_traces(
                vector_observations, legend_group=vector_name
            )
        }
        for ensemble in self._selected_ensembles:
            self._add_vector_traces_set_to_figure(
                vector_observations_traces_set, ensemble
            )

    #############################################################################
    #
    # Private methods
    #
    #############################################################################

    def _set_keep_uirevision(
        self,
    ) -> None:
        # Keep uirevision (e.g. zoom) for unchanged data.
        self._figure.update_xaxes(uirevision="locked")  # Time axis state kept
        for i, owner in enumerate(self._selected_ensembles, start=1):
            self._figure.update_yaxes(row=i, col=1, uirevision=owner)

    def _add_vector_trace_set_to_figure(
        self, vector_trace_set: Dict[str, dict], ensemble: Optional[str] = None
    ) -> None:
        for trace in vector_trace_set.values():
            subplot_index = (
                self._selected_ensembles.index(ensemble) + 1
                if ensemble in self._selected_ensembles
                else None
            )
            if subplot_index is None:
                continue
            self._figure.add_trace(trace, row=subplot_index, col=1)

    def _add_vector_traces_set_to_figure(
        self, vector_traces_set: Dict[str, List[dict]], ensemble: Optional[str] = None
    ) -> None:
        for vector_traces in vector_traces_set.values():
            subplot_index = (
                self._selected_ensembles.index(ensemble) + 1
                if ensemble in self._selected_ensembles
                else None
            )
            if subplot_index is None:
                continue
            self._figure.add_traces(vector_traces, rows=subplot_index, cols=1)

    def _validate_vectors_are_selected(self, vectors: Set[str]) -> None:
        """Validate set of vectors are among selected vectors

        Check if vectors are among selected vectors for figure builder, raise
        ValueError if not.

        `Input:`
        * vectors: Set[str] - set of vector names to verify
        """
        for vector in vectors:
            if vector not in self._selected_vectors:
                raise ValueError(
                    f'Vector "{vector}" does not exist among selected vectors: '
                    f"{self._selected_vectors}"
                )



from typing import Dict, List, Tuple

import webviz_core_components as wcc
from dash import Input, Output, State, callback
from dash.development.base_component import Component
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._utils import RftPlotterDataModel


class Selections(SettingsGroupABC):
    class Ids(StrEnum):
        ENSEMBLE = "param-response-ensemble"
        WELL = "param-response-well"
        DATE = "param-response-date"
        ZONE = "param-response-zone"
        PARAM = "param-response-param"

    def __init__(self, datamodel: RftPlotterDataModel) -> None:
        super().__init__("Selections")
        self._datamodel = datamodel
        self._ensembles = datamodel.ensembles
        self._well_names = datamodel.well_names
        self._params = datamodel.parameters if not datamodel.parameters is None else []
        self._parameter_df = datamodel.param_model.dataframe

        well = self._well_names[0] if self._well_names else ""

        self._dates_in_well, self._zones_in_well = self._datamodel.well_dates_and_zones(
            well
        )

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                label="Ensemble",
                id=self.register_component_unique_id(self.Ids.ENSEMBLE),
                options=[{"label": ens, "value": ens} for ens in self._ensembles],
                value=self._ensembles[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Well",
                id=self.register_component_unique_id(self.Ids.WELL),
                options=[{"label": well, "value": well} for well in self._well_names],
                value=self._well_names[0] if self._well_names else "",
                clearable=False,
            ),
            wcc.Dropdown(
                label="Date",
                id=self.register_component_unique_id(self.Ids.DATE),
                options=[
                    {"label": date, "value": date} for date in self._dates_in_well
                ],
                value=self._dates_in_well[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Zone",
                id=self.register_component_unique_id(self.Ids.ZONE),
                options=[
                    {"label": zone, "value": zone} for zone in self._zones_in_well
                ],
                value=self._zones_in_well[0],
                clearable=False,
            ),
            wcc.Dropdown(
                label="Parameter",
                id=self.register_component_unique_id(self.Ids.PARAM),
                options=[{"label": param, "value": param} for param in self._params],
                clearable=False,
                value=None,
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(self.component_unique_id(self.Ids.DATE).to_string(), "options"),
            Output(self.component_unique_id(self.Ids.DATE).to_string(), "value"),
            Output(self.component_unique_id(self.Ids.ZONE).to_string(), "options"),
            Output(self.component_unique_id(self.Ids.ZONE).to_string(), "value"),
            Input(self.component_unique_id(self.Ids.WELL).to_string(), "value"),
            State(self.component_unique_id(self.Ids.ZONE).to_string(), "value"),
        )
        @callback_typecheck
        def _update_date_and_zone(
            well: str, zone_state: str
        ) -> Tuple[List[Dict[str, str]], str, List[Dict[str, str]], str]:
            """Update dates and zones when selecting well. If the current
            selected zone is also present in the new well it will be kept as value.
            """

            dates_in_well, zones_in_well = self._datamodel.well_dates_and_zones(well)
            return (
                [{"label": date, "value": date} for date in dates_in_well],
                dates_in_well[0],
                [{"label": zone, "value": zone} for zone in zones_in_well],
                zone_state if zone_state in zones_in_well else zones_in_well[0],
            )



import pathlib

import pandas as pd
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: pathlib.Path) -> pd.DataFrame:
    return pd.read_csv(csv_file)



from collections.abc import Iterable
from typing import Dict, List, Optional

import numpy as np
import pandas as pd


def filter_and_sum_responses(
    dframe: pd.DataFrame,
    ensemble: str,
    response: str,
    filteroptions: Optional[List[Dict[str, str]]] = None,
    aggregation: str = "sum",
) -> pd.DataFrame:
    """Filter response dataframe for the given ensemble
    and optional filter columns. Returns dataframe grouped and
    aggregated per realization.
    """

    df = dframe.loc[dframe["ENSEMBLE"] == ensemble]
    if filteroptions:
        for opt in filteroptions:
            if opt["type"] == "multi" or opt["type"] == "single":
                if isinstance(opt["values"], list):
                    df = df.loc[df[opt["name"]].isin(opt["values"])]
                else:
                    if opt["name"] == "DATE" and isinstance(opt["values"], str):
                        df = df.loc[df["DATE"].astype(str) == opt["values"]]
                    else:
                        df = df.loc[df[opt["name"]] == opt["values"]]

            elif opt["type"] == "range":
                df = df.loc[
                    (df[opt["name"]] >= np.min(opt["values"]))
                    & (df[opt["name"]] <= np.max(opt["values"]))
                ]
    if aggregation == "sum":
        return df.groupby("REAL").sum().reset_index()[["REAL", response]]
    if aggregation == "mean":
        return df.groupby("REAL").mean().reset_index()[["REAL", response]]
    raise ValueError(f"Unknown aggregation '{aggregation}'.")


def check_runs(parameterdf: pd.DataFrame, responsedf: pd.DataFrame) -> None:
    """Check that input parameters and response files have
    the same number of runs"""
    for col in ["ENSEMBLE", "REAL"]:
        if sorted(list(parameterdf[col].unique())) != sorted(
            list(responsedf[col].unique())
        ):
            raise ValueError("Parameter and response files have different runs")


def check_response_filters(
    responsedf: pd.DataFrame, response_filters: Optional[dict]
) -> None:
    """Check that provided response filters are valid"""
    if response_filters:
        for col_name, col_type in response_filters.items():
            if col_name not in responsedf.columns:
                raise ValueError(f"{col_name} is not in response file")
            if col_type not in ["single", "multi", "range"]:
                raise ValueError(f"Filter type {col_type} for {col_name} is not valid.")


def filter_numerical_columns(
    df: pd.DataFrame,
    column_ignore: Optional[List[str]] = None,
    column_include: Optional[List[str]] = None,
    filter_columns: Optional[Iterable] = None,
) -> List[str]:
    """Filter to numerical columns, and respect ignore/include and filters.
    Also remove ENSEMBLE and REAL. Return list of column labels."""
    cols: List[str] = []
    for col in df.columns[[np.issubdtype(dtype, np.number) for dtype in df.dtypes]]:
        if (
            (column_ignore and col in column_ignore)
            or (col in ["ENSEMBLE", "REAL"])
            or (filter_columns and col in filter_columns)
        ):
            continue
        if column_include and col in column_include or column_include is None:
            cols.append(col)
    return cols


def make_response_filters(
    response_filters: dict, response_filter_values: list
) -> List[dict]:
    """Returns a list of active response filters"""
    filteroptions = []
    if response_filter_values:
        for i, (col_name, col_type) in enumerate(response_filters.items()):
            filteroptions.append(
                {
                    "name": col_name,
                    "type": col_type,
                    "values": response_filter_values[i],
                }
            )
    return filteroptions



import glob
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from fmu.ensemble import ScratchEnsemble
except ImportError:
    pass


@dataclass(frozen=True)
class FaultPolygonsFileInfo:
    path: str
    real: int
    name: str
    attribute: str


def _discover_ensemble_realizations_fmu(ens_path: str) -> Dict[int, str]:
    """Returns dict indexed by realization number and with runpath as value"""
    scratch_ensemble = ScratchEnsemble("dummyEnsembleName", paths=ens_path).filter("OK")
    real_dict = {i: r.runpath() for i, r in scratch_ensemble.realizations.items()}
    return real_dict


def _discover_ensemble_realizations(ens_path: str) -> Dict[int, str]:
    # Much faster than FMU impl above, but is it risky?
    # Do we need to check for OK-file?
    real_dict: Dict[int, str] = {}

    realidxregexp = re.compile(r"realization-(\d+)")
    globbed_real_dirs = sorted(glob.glob(str(ens_path)))
    for real_dir in globbed_real_dirs:
        realnum: Optional[int] = None
        for path_comp in reversed(real_dir.split(os.path.sep)):
            realmatch = re.match(realidxregexp, path_comp)
            if realmatch:
                realnum = int(realmatch.group(1))
                break

        if realnum is not None:
            real_dict[realnum] = real_dir

    return real_dict


@dataclass(frozen=True)
class FaultPolygonsIdent:
    name: str
    attribute: str


def _fault_polygons_ident_from_filename(filename: str) -> Optional[FaultPolygonsIdent]:
    """Split the stem part of the fault polygons filename into fault polygons name and attribute"""
    delimiter: str = "--"
    parts = Path(filename).stem.split(delimiter)
    if len(parts) != 2:
        return None

    return FaultPolygonsIdent(name=parts[0], attribute=parts[1])


def discover_per_realization_fault_polygons_files(
    ens_path: str,
) -> List[FaultPolygonsFileInfo]:
    rel_fault_polygons_folder: str = "share/results/polygons"
    suffix: str = "*.pol"

    fault_polygons_files: List[FaultPolygonsFileInfo] = []

    real_dict = _discover_ensemble_realizations_fmu(ens_path)
    for realnum, runpath in sorted(real_dict.items()):
        globbed_filenames = glob.glob(
            str(Path(runpath) / rel_fault_polygons_folder / suffix)
        )
        for fault_polygons_filename in sorted(globbed_filenames):
            fault_polygons_ident = _fault_polygons_ident_from_filename(
                fault_polygons_filename
            )
            if fault_polygons_ident:
                fault_polygons_files.append(
                    FaultPolygonsFileInfo(
                        path=fault_polygons_filename,
                        real=realnum,
                        name=fault_polygons_ident.name,
                        attribute=fault_polygons_ident.attribute,
                    )
                )

    return fault_polygons_files
